from typing import Dict, List

import pytest  # pyre-ignore
from fastapi import status

from karp import auth
from karp.errors import ClientErrorCodes
from karp.lex.application.queries.resources import GetEntryRepositoryId
from karp.lex.application.repositories.entry_repositories import EntryUowRepositoryUnitOfWork
from karp.utility.time import utc_now
from karp.lex.application.queries import EntryDto

# from karp.application import ctx, config
# from karp.infrastructure.unit_of_work import unit_of_work


# from tests.utils import get_json

# from tests.integration_tests.common_fixtures import (
#     fixture_fa_data_client,
#     fixture_places,
# )

def get_entry_uow(container, resource_id: str):
    get_entry_repository_id = container.get(GetEntryRepositoryId)
    entry_repo_id = get_entry_repository_id.query('places')
    entry_uow_repo_uow = container.get(EntryUowRepositoryUnitOfWork)
    with entry_uow_repo_uow as uw:
        return uw.repo.get_by_id(entry_repo_id)


def init(
    client,
    entries: List[Dict],
    access_token: auth.AccessToken,
):

    for entry in entries:
        response = client.post(
            '/entries/places/add',
            json={'entry': entry},
            headers=access_token.as_header(),
        )
        assert response.status_code < 300, response.status_code
    return client


class TestEntriesRoutes:
    def test_routes_exist(self, fa_data_client):
        response = fa_data_client.post('/entries/places/add')
        assert response.status_code != status.HTTP_404_NOT_FOUND

        response = fa_data_client.post('/entries/places/update')
        assert response.status_code != status.HTTP_404_NOT_FOUND

        response = fa_data_client.put('/entries/places')
        assert response.status_code != status.HTTP_404_NOT_FOUND

        response = fa_data_client.post('/entries/places/preview')
        assert response.status_code != status.HTTP_404_NOT_FOUND


class TestAddEntry:
    def test_put_route_exist(self, fa_data_client):
        response = fa_data_client.put('/entries/places')
        assert response.status_code != status.HTTP_404_NOT_FOUND

    @pytest.mark.parametrize('invalid_data', [
        ({},),
        ({'user': 'a@b.se'},),
    ])
    def test_invalid_data_returns_422(
        self,
        fa_data_client,
        invalid_data: Dict,
        write_token: auth.AccessToken,
    ):
        response = fa_data_client.post(
            '/entries/places/add',
            json=invalid_data,
            headers=write_token.as_header(),
        )
        print(f'{response.json()=}')
        assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

    def test_add_with_valid_data_returns_201(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):

        response = fa_data_client.put(
            '/entries/places',
            json={
                'entry': {
                    'code': 203,
                    'name': 'add203',
                    'population': 4,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        print(f'response. = {response.json()}')
        assert response.status_code == 201
        response_data = response.json()
        assert 'newID' in response_data
        assert response_data['newID'] == '203'

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container,
            resource_id='places'
        )
        with entry_uow as uw:
            assert '203' in uw.repo.entry_ids()

    def test_adding_existing_fails_with_400(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 204
        entry_name = f'add{entry_id}'
        response = fa_data_client.post(
            '/entries/places/add',
            json={
                'entry': {
                    'code': entry_id,
                    'name': entry_name,
                    'population': 4,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        print(f'response = {response.json()}')

        assert response.status_code == 201

        response = fa_data_client.post(
            '/entries/places/add',
            json={
                'entry': {
                    'code': entry_id,
                    'name': entry_name,
                    'population': 4,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        assert response.status_code == 400
        response_data = response.json()

        assert 'error' in response_data
        assert 'errorCode' in response_data
        assert ClientErrorCodes.DB_INTEGRITY_ERROR == response_data['errorCode']
        assert (
            response_data['error']
            == f"An entry with entry_id '{entry_id}' already exists."
        )

    def test_add_fails_with_invalid_entry(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        response = fa_data_client.post(
            '/entries/places/add',
            json={'entry': {}},
            headers=write_token.as_header()
        )

        assert response.status_code == status.HTTP_400_BAD_REQUEST
        response_data = response.json()

        assert response_data['error'] == "Missing ID field for resource 'places' in '{}'"
        assert response_data['errorCode'] == ClientErrorCodes.ENTRY_NOT_VALID


class TestDeleteEntry:
    def test_delete(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 205
        entry_name = f'delete{entry_id}'
        response = fa_data_client.post(
            '/entries/places/add',
            json={
                'entry': {
                    'code': entry_id,
                    'name': entry_name,
                    'population': 4,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                }
            },
            headers=write_token.as_header(),
        )

        response = fa_data_client.delete(
            f'/entries/places/{entry_id}/delete',
            headers=write_token.as_header()
        )

        assert response.status_code == status.HTTP_200_OK

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container,
            resource_id='places'
        )
        with entry_uow as uw:
            assert uw.repo.by_entry_id(str(entry_id)).discarded
            assert str(entry_id) not in uw.repo.entry_ids()

    def test_delete_non_existing_fails(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):

        entry_id = 'non_existing_id'

        response = fa_data_client.delete(
            f'/entries/places/{entry_id}/delete',
            headers=write_token.as_header()
        )

        assert response.status_code == status.HTTP_404_NOT_FOUND

        response_data = response.json()

        assert response_data['errorCode'] == ClientErrorCodes.ENTRY_NOT_FOUND

        assert (
            response_data['error']
            == f"Entry '{entry_id}' not found in resource 'places' (version=latest)"
        )


class TestUpdateEntry:
    def test_update_non_existing_fails(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 'non-existent'
        response = fa_data_client.post(
            f'/entries/places/{entry_id}/update',
            json={
                'entry': {
                    'code': 3,
                    'name': 'test3',
                    'population': 5,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                },
                'message': 'changes',
                'version': 1,
            },
            headers=write_token.as_header(),
        )
        assert response.status_code == status.HTTP_404_NOT_FOUND
        response_data = response.json()
        assert response_data['errorCode'] == ClientErrorCodes.ENTRY_NOT_FOUND

        assert (
            response_data['error']
            == f"Entry '{entry_id}' not found in resource 'places' (version=latest)"
        )

    def test_update_wo_changes_succeeds(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 206
        entry_name = f'update{entry_id}'
        response = fa_data_client.post(
            '/entries/places/add',
            json={
                'entry': {
                    'code': entry_id,
                    'name': entry_name,
                    'population': 4,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        print(f'response = {response.json()}')

        assert response.status_code == status.HTTP_201_CREATED

        response = fa_data_client.post(
            f'/entries/places/{entry_id}/update',
            json={
                'entry': {
                    'code': entry_id,
                    'name': entry_name,
                    'population': 4,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                },
                'message': 'changes',
                'version': 1,
            },
            headers=write_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK

    def test_update_wrong_version_fails(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 207
        response = fa_data_client.post(
            '/entries/places/add',
            json={
                'entry': {
                    'code': entry_id,
                    'name': 'update3',
                    'population': 4,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        print(f'response = {response.json()}')

        assert response.status_code == 201

        response = fa_data_client.post(
            f'/entries/places/{entry_id}/update',
            json={
                'entry': {
                    'code': entry_id,
                    'name': 'update3',
                    'population': 5,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                },
                'message': 'changes',
                'version': 2,
            },
            headers=write_token.as_header(),
            #         ),
            #         content_type='application/json',
        )
        assert response.status_code == 400
        response_data = response.json()
        assert response_data['errorCode'] == ClientErrorCodes.VERSION_CONFLICT

        assert response_data['error'] == 'Version conflict. Please update entry.'
        assert response_data['diff'] == [
            {'type': 'CHANGE', 'field': 'population', 'before': 4, 'after': 5}
        ]

    def test_update_returns_200_on_valid_data(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 208
        entry_name = f'update{entry_id}'
        response = fa_data_client.post(
            '/entries/places/add',
            json={
                'entry': {
                    'code': entry_id,
                    'name': entry_name,
                    'population': 4,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        print(f'response = {response.json()}')

        assert response.status_code == 201

        response = fa_data_client.post(
            f'/entries/places/{entry_id}/update',
            json={
                'entry': {
                    'code': entry_id,
                    'name': entry_name,
                    'population': 5,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                },
                'message': 'changes',
                'version': 1,
            },
            headers=write_token.as_header(),
        )
        assert response.status_code == 200
        response_data = response.json()
        assert response_data['newID'] == str(entry_id)

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container,
            resource_id='places')
        with entry_uow as uw:
            assert uw.repo.by_entry_id(str(entry_id)).body['population'] == 5
            assert str(entry_id) in uw.repo.entry_ids()

    def test_update_several_times(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 209
        response = fa_data_client.post(
            '/entries/places/add',
            json={
                'entry': {
                    'code': entry_id,
                    'name': 'a',
                    'municipality': [1]
                }
            },
            headers=write_token.as_header(),
        )
        assert response.status_code == status.HTTP_201_CREATED

        for i in range(2, 10):
            response = fa_data_client.post(
                f'/entries/places/{entry_id}/update',
                json={
                    'entry': {
                        'code': entry_id,
                        'name': 'a' * i,
                        'municipality': [1]
                    },
                    'message': 'changes',
                    'version': i - 1,
                },
                headers=write_token.as_header(),
            )
            print(f'i = {i}: response = {response.json()}')
            assert response.status_code == status.HTTP_200_OK

    @pytest.mark.xfail()
    def test_update_entry_id(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 210
        response = fa_data_client.post(
            '/entries/places/add',
            json={
                'entry': {
                    'code': entry_id,
                    'name': 'update3',
                    'population': 4,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                }
            },
            headers=write_token.as_header(),
        )
        assert response.status_code == 201

        response = fa_data_client.post(
            f'places/{entry_id}/update',
            json={
                'entry': {
                    'code': entry_id + 1,
                    'name': 'update3',
                    'population': 5,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                },
                'message': 'changes',
                'version': 1,
            },
            headers=write_token.as_header(),
        )
        response_data = response.json()
        print(f'{response.json()=}')
        assert str(entry_id + 1) == response_data['newID']

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container,
            resource_id='places'
        )
        with entry_uow as uw:
            entry_ids = uw.repo.entry_ids()
            assert str(entry_id) not in entry_ids
            assert str(entry_id + 1) in entry_ids

    def test_update_changes_last_modified(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        entry_id = 212
        before_add = utc_now()

        init(
            fa_data_client,
            [
                {
                    'code': entry_id,
                    'name': 'last_modified1',
                    'municipality': [1]
                },
            ],
            write_token
        )

        after_add = utc_now()

        entry_uow = get_entry_uow(
            fa_data_client.app.state.app_context.container,
            resource_id='places'
        )
        with entry_uow as uw:
            entry = uw.repo.by_entry_id(str(entry_id))
            assert entry.last_modified > before_add
            assert entry.last_modified < after_add

        fa_data_client.post(
            f'/entries/places/{entry_id}/update',
            json={
                'entry': {
                    'code': entry_id,
                    'name': 'last_modified2',
                    'municipality': [1]
                },
                'message': 'changes',
                'version': 1,
            },
            headers=write_token.as_header(),
        )

        after_update = utc_now()

        with entry_uow as uw:
            entry = uw.repo.by_entry_id(str(entry_id))
            assert entry.last_modified > after_add
            assert entry.last_modified < after_update


class TestGetEntry:
    def test_get_entry_wo_auth_returns_403(self, fa_data_client):
        response = fa_data_client.get('/entries/places/204')
        assert response.status_code == status.HTTP_403_FORBIDDEN

    def test_get_entry_w_lower_auth_returns_401(
        self,
        fa_data_client,
        write_token: auth.AccessToken,
    ):
        response = fa_data_client.get(
            '/entries/places/204',
            headers=write_token.as_header(),
        )
        assert response.status_code == status.HTTP_401_UNAUTHORIZED

    def test_get_entry_by_entry_id(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
    ):
        response = fa_data_client.get(
            '/entries/places/204',
            headers=admin_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK

        entry = EntryDto(**response.json())
        assert entry.entry_id == '204'
        assert entry.version == 1

    def test_route_w_version_exist(
        self,
        fa_data_client,
        admin_token: auth.AccessToken,
    ):
        response = fa_data_client.get(
            '/entries/places/209/5',
            headers=admin_token.as_header(),
        )
        assert response.status_code == status.HTTP_200_OK

        entry = EntryDto(**response.json())
        assert entry.entry_id == '209'
        assert entry.version == 5


class TestPreviewEntry:
    def test_preview_fails_with_422_on_invalid_data(self, fa_data_client, read_token: auth.AccessToken):
        response = fa_data_client.post(
            '/entries/places/preview', json={}, headers=read_token.as_header())
        assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

    def test_preview_returns_200_on_valid_data(self, fa_data_client, read_token: auth.AccessToken):
        response = fa_data_client.post(
            '/entries/places/preview',
            json={
                'entry': {
                    'code': 3,
                    'name': 'update3',
                    'population': 4,
                    'area': 50000,
                    'density': 5,
                    'municipality': [2, 3],
                },
                'message': 'test'
            },
            headers=read_token.as_header()
        )
        print(f'{response.json()=}')
        assert response.status_code == status.HTTP_200_OK
        assert response.json()['entry']['id'] == '3'


@pytest.mark.skip()
def test_update_wrong_id(fa_data_client, write_token: auth.AccessToken):
    response = fa_data_client.post(
        '/entries/places/add',
        json={
            'entry': {
                'code': 3,
                'name': 'update3',
                'population': 4,
                'area': 50000,
                'density': 5,
                'municipality': [2, 3],
            }
        },
        headers=write_token.as_header(),
    )
    print(f'response = {response.json()}')

    assert response.status_code == 201
    entry_id = response.json()['newID']

    with unit_of_work(using=ctx.resource_repo) as uw:
        resource = uw.get_active_resource('places')

    with unit_of_work(using=resource.entry_repository) as uw:
        entries = uw.entry_ids()
        assert len(entries) == 1
        assert entries[0] == '3'

    response = fa_data_client.post(
        f'places/{entry_id}/update',
        json={
            'entry': {
                'code': 4,
                'name': 'update3',
                'population': 4,
                'area': 50000,
                'density': 5,
                'municipality': [2, 3],
            },
            'message': 'changes',
            'version': 1,
        },
        headers=write_token.as_header(),
    )
    assert response.status_code == 400
    response_data = response.json()
    assert response_data['errorCode'] == ClientErrorCodes.ENTRY_ID_MISMATCH

    assert response_data['error'] == "entry_id '4' does not equal '3'"


@pytest.mark.skip()
def test_refs(fa_data_client):
    client = init(
        fa_data_client,
        [
            {
                'code': 1,
                'name': 'refs1refs',
                'population': 10,
                'area': 50000,
                'density': 5,
                'municipality': [2, 3],
            },
            {
                'code': 2,
                'name': 'refs2refs',
                'population': 5,
                'larger_place': 1,
                'area': 50000,
                'density': 5,
                'municipality': [2, 3],
            },
        ],
    )

    entries = get_json(client, 'places/query')
    assert len(entries['hits']) == 2
    for val in entries['hits']:
        assert 'entry' in val
        entry = val['entry']
        print('entry = {}'.format(entry))
        if entry['code'] == 1:
            assert 'v_larger_place' not in entry
            assert 'larger_place' not in entry
            assert 'v_smaller_places' in entry
            assert entry['v_smaller_places'][0]['code'] == 2
        else:
            assert entry['v_larger_place']['code'] == 1
            assert entry['v_larger_place']['name'] == 'refs1refs'
            assert 'v_smaller_places' not in entry


@pytest.mark.skip()
def test_external_refs(fa_data_client):
    client = init(
        fa_data_client,
        [
            {
                'code': 1,
                'name': 'test1',
                'population': 10,
                'area': 50000,
                'density': 5,
                'municipality': [1],
            },
            {
                'code': 2,
                'name': 'test2',
                'population': 5,
                'larger_place': 1,
                'area': 50000,
                'density': 5,
                'municipality': [1, 2],
            },
            {
                'code': 3,
                'name': 'test2',
                'population': 5,
                'larger_place': 1,
                'area': 50000,
                'density': 5,
                'municipality': [2],
            },
        ],
    )

    client.post(
        'municipalities/add',
        data=json.dumps(
            {
                'entry': {
                    'code': 1,
                    'name': 'municipality1',
                    'state': 'state1',
                    'region': 'region1',
                }
            }
        ),
        content_type='application/json',
    )

    client.post(
        'municipalities/add',
        data=json.dumps(
            {
                'entry': {
                    'code': 2,
                    'name': 'municipality2',
                    'state': 'state2',
                    'region': 'region2',
                }
            }
        ),
        content_type='application/json',
    )

    entries = get_json(client, 'municipalities/query')
    for val in entries['hits']:
        assert 'entry' in val
        entry = val['entry']

        assert 'v_places' in entry
        place_codes = [place['code'] for place in entry['v_places']]
        assert len(place_codes) == 2
        if entry['code'] == 1:
            assert 1 in place_codes
            assert 2 in place_codes
        else:
            assert 2 in place_codes
            assert 3 in place_codes

    places_entries = get_json(client, 'places/query')
    for val in places_entries['hits']:
        assert 'entry' in val
        entry = val['entry']
        assert 'municipality' in entry
        assert isinstance(entry['v_municipality'], list)
        if entry['code'] == 2:
            assert {'code': 1, 'name': 'municipality1', 'state': 'state1'} in entry[
                'v_municipality'
            ]
            assert {'code': 2, 'name': 'municipality2', 'state': 'state2'} in entry[
                'v_municipality'
            ]


@pytest.mark.skip()
def test_update_refs(fa_data_client):
    client = init(
        fa_data_client,
        es,
        [
            {
                'code': 5,
                'name': 'test1',
                'population': 10,
                'area': 50000,
                'density': 5,
                'municipality': [2, 3],
            },
            {
                'code': 6,
                'name': 'test2',
                'population': 5,
                'larger_place': 5,
                'area': 50000,
                'density': 5,
                'municipality': [2, 3],
            },
        ],
    )

    entries = get_json(client, 'places/query')
    assert len(entries['hits']) == 2
    for val in entries['hits']:
        assert 'entry' in val
        entry = val['entry']
        print('entry = {}'.format(entry))
        if entry['code'] == 5:
            assert 'v_smaller_places' in entry
            assert entry['v_smaller_places'][0]['code'] == 6

    client.delete('/places/6/delete')

    entries = get_json(client, 'places/query')
    assert len(entries['hits']) == 1
    entry = entries['hits'][0]
    assert 'v_smaller_places' not in entry


@pytest.mark.skip()
def test_update_refs2(fa_data_client):
    client = init(fa_data_client, es, [
                  {'code': 3, 'name': 'test3', 'municipality': [2, 3]}])

    client.post(
        'places/3/update',
        data=json.dumps(
            {
                'entry': {'code': 3, 'name': 'test3', 'municipality': [2]},
                'message': 'changes',
                'version': 1,
            }
        ),
        content_type='application/json',
    )

    entries = get_json(client, 'places/query')
    assert len(entries['hits']) == 1
    assert entries['hits'][0]['id'] == '3'
    assert entries['hits'][0]['entry']['municipality'] == [2]
    assert (
        'v_municipality' not in entries['hits'][0]
        or len(entries['hits'][0]['municipality']) == 0
    )
    with client.application.app_context():
        db_entry = entryread.get_entry('places', '3')
        assert len(db_entry.municipality) == 1
        assert db_entry.municipality[0].municipality == 2



from dependency_injector import containers, providers

from karp.infrastructure.jwt import jwt_auth_service
from karp.main.containers import AppContainer


class WebAppContainer(containers.DeclarativeContainer):

    config = providers.Configuration()

    context = providers.Container(
        AppContainer,
        config=config,
    )

    jwt_authenticator = providers.Singleton(
        jwt_auth_service.JWTAuthenticator,
        pubkey_path=config.auth.jwt.pubkey_path,
        resource_uow=context.resource_uow,
    )

    auth_service = providers.Selector(
        config.auth.type,
        jwt_auth=jwt_authenticator
    )



import typing

from karp.foundation.value_objects.permission_level import PermissionLevel


class User:
    def __init__(
        self,
        identifier: str,
        permissions: typing.Dict[str, int],
        levels: typing.Dict[str, int]
    ):
        self._identifier = identifier
        self._permissions = permissions
        self._levels = levels

    @property
    def identifier(self) -> str:
        return self._identifier

    def has_enough_permissions(self, resource_id: str, level: PermissionLevel) -> bool:
        return self._permissions.get(resource_id) and (self._permissions[resource_id] >= self._levels[level])



import logging
from typing import Optional

from fastapi import (APIRouter, Body, Depends, HTTPException, Response, Security,
                     status, Path, Query)
import pydantic
from starlette import responses
import logging

from karp import errors as karp_errors, auth, lex, search
from karp.lex.application.queries import EntryViews, EntryDto, GetEntryHistory
from karp.lex.domain import commands, errors
from karp.auth import User
from karp.foundation.value_objects import PermissionLevel, unique_id
from karp.auth import AuthService
from karp.webapp import schemas, dependencies as deps

from karp.webapp.dependencies.fastapi_injector import inject_from_req

router = APIRouter()

logger = logging.getLogger(__name__)


@router.get('/{resource_id}/{entry_id}/{version}', response_model=EntryDto, tags=["History"])
@router.get('/{resource_id}/{entry_id}', response_model=EntryDto, tags=["History"])
# @auth.auth.authorization("ADMIN")
def get_history_for_entry(
    resource_id: str,
    entry_id: str,
    version: Optional[int] = Query(None),
    user: auth.User = Security(deps.get_user, scopes=["admin"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    get_entry_history: GetEntryHistory = Depends(deps.get_entry_history),
):
    if not auth_service.authorize(
        auth.PermissionLevel.admin, user, [resource_id]
    ):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="lexica:admin"'},
        )
    logger.info('getting history for entry', extra={'resource_id': resource_id,
                                                    'entry_id': entry_id, 'user': user.identifier})
    historical_entry = get_entry_history.query(
        resource_id, entry_id, version=version
    )

    return historical_entry


@ router.post("/{resource_id}/add", status_code=status.HTTP_201_CREATED, tags=["Editing"])
@ router.put('/{resource_id}', status_code=status.HTTP_201_CREATED, tags=["Editing"])
def add_entry(
    resource_id: str,
    data: schemas.EntryAdd,
    user: User = Security(deps.get_user, scopes=["write"]),
    auth_service: AuthService = Depends(deps.get_auth_service),
    adding_entry_uc: lex.AddingEntry = Depends(
        deps.get_lex_uc(lex.AddingEntry)),
):

    if not auth_service.authorize(PermissionLevel.write, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="write"'},
        )
    logger.info('adding entry', extra={
                'resource_id': resource_id, 'data': data})
    try:
        new_entry = adding_entry_uc.execute(
            commands.AddEntry(
                resource_id=resource_id,
                user=user.identifier,
                message=data.message,
                entry=data.entry,
            )
        )
    except errors.IntegrityError as exc:
        return responses.JSONResponse(
            status_code=400,
            content={
                'error': str(exc),
                'errorCode': karp_errors.ClientErrorCodes.DB_INTEGRITY_ERROR
            }
        )
    except errors.InvalidEntry as exc:
        return responses.JSONResponse(
            status_code=400,
            content={
                'error': str(exc),
                'errorCode': karp_errors.ClientErrorCodes.ENTRY_NOT_VALID
            }
        )

    return {"newID": new_entry.entry_id, "entityID": new_entry.entity_id}


@ router.post('/{resource_id}/preview')
# @auth.auth.authorization("READ")
def preview_entry(
    resource_id: str,
    data: schemas.EntryAdd,
    user: auth.User = Security(deps.get_user_optional, scopes=['read']),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    preview_entry: search.PreviewEntry = Depends(
        deps.inject_from_req(search.PreviewEntry)),
):
    if not auth_service.authorize(PermissionLevel.read, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={'WWW-Authenticate': 'Bearer scope="lexica:read"'}
        )
    try:
        input_dto = search.PreviewEntryInputDto(
            resource_id=resource_id,
            entry=data.entry,
            user=user.identifier
        )
    except pydantic.ValidationError as err:
        logger.exception('data is not valid')
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail={'error': str(err)})
    else:
        return preview_entry.query(input_dto)
#     data = request.get_json()
#     preview = entrywrite.preview_entry(resource_id, data)
#     return flask_jsonify(preview)


@ router.post("/{resource_id}/{entry_id}/update", tags=["Editing"])
@ router.post('/{resource_id}/{entry_id}', tags=["Editing"])
# @auth.auth.authorization("WRITE", add_user=True)
def update_entry(
    response: Response,
    resource_id: str,
    entry_id: str,
    data: schemas.EntryUpdate,
    user: User = Security(deps.get_user, scopes=["write"]),
    auth_service: AuthService = Depends(deps.get_auth_service),
    updating_entry_uc: lex.UpdatingEntry = Depends(
        deps.get_lex_uc(lex.UpdatingEntry)),
):
    if not auth_service.authorize(PermissionLevel.write, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="write"'},
        )

    #     force_update = convert.str2bool(request.args.get("force", "false"))
    #     data = request.get_json()
    #     version = data.get("version")
    #     entry = data.get("entry")
    #     message = data.get("message")
    #     if not (version and entry and message):
    #         raise KarpError("Missing version, entry or message")
    logger.info('updating entry', extra={'resource_id': resource_id,
                                         'entry_id': entry_id, 'data': data, 'user': user.identifier})
    try:
        entry = updating_entry_uc.execute(
            commands.UpdateEntry(
                resource_id=resource_id,
                entry_id=entry_id,
                version=data.version,
                user=user.identifier,
                message=data.message,
                entry=data.entry,
            )
        )
        # new_entry = entries.add_entry(
        #     resource_id, data.entry, user.identifier, message=data.message
        # )
        # new_id = entries.update_entry(
        #     resource_id,
        #     entry_id,
        #     data.version,
        #     data.entry,
        #     user.identifier,
        #     message=data.message,
        #     # force=force_update,
        # )
        return {"newID": entry.entry_id, "entityID": entry.entity_id}
    except errors.EntryNotFound:
        return responses.JSONResponse(
            status_code=404,
            content={
                'error': f"Entry '{entry_id}' not found in resource '{resource_id}' (version=latest)",
                'errorCode': karp_errors.ClientErrorCodes.ENTRY_NOT_FOUND,
                'resource': resource_id,
                'entry_id': entry_id,
            },
        )
    except errors.UpdateConflict as err:
        response.status_code = status.HTTP_400_BAD_REQUEST
        err.error_obj["errorCode"] = karp_errors.ClientErrorCodes.VERSION_CONFLICT
        return err.error_obj
    except Exception as err:
        logger.exception('error occured', extra={'resource_id': resource_id,
                         'entry_id': entry_id, 'data': data})
        raise


@ router.delete('/{resource_id}/{entry_id}/delete', tags=["Editing"])
@ router.delete('/{resource_id}/{entry_id}', tags=["Editing"], status_code=status.HTTP_204_NO_CONTENT)
# @auth.auth.authorization("WRITE", add_user=True)
def delete_entry(
    resource_id: str,
    entry_id: str,
    user: User = Security(deps.get_user, scopes=["write"]),
    auth_service: AuthService = Depends(deps.get_auth_service),
    deleting_entry_uc: lex.DeletingEntry = Depends(
        deps.get_lex_uc(lex.DeletingEntry))
):
    """Delete a entry from a resource.

    Arguments:
        user {karp.auth.user.User} -- [description]
        resource_id {str} -- [description]
        entry_id {str} -- [description]

    Returns:
        [type] -- [description]
    """
    if not auth_service.authorize(PermissionLevel.write, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="lexica:write"'},
        )
    try:
        deleting_entry_uc.execute(
            commands.DeleteEntry(
                resource_id=resource_id,
                entry_id=entry_id,
                user=user.identifier,
            )
        )
    except errors.EntryNotFound:
        return responses.JSONResponse(
            status_code=404,
            content={
                'error': f"Entry '{entry_id}' not found in resource '{resource_id}' (version=latest)",
                'errorCode': karp_errors.ClientErrorCodes.ENTRY_NOT_FOUND,
                'resource': resource_id,
                'entry_id': entry_id,
            }
        )
    # entries.delete_entry(resource_id, entry_id, user.identifier)
    return


def init_app(app):
    app.include_router(router)



import injector

from karp.foundation.commands import CommandHandler
from karp.lex.application.repositories import (
    EntryUowRepositoryUnitOfWork,
    EntryRepositoryUnitOfWorkFactory,
    InjectorEntryUnitOfWorkRepoFactory,
    ResourceUnitOfWork,
)
from karp.lex.domain.commands import (
    CreateEntryRepository,
    CreateResource,
)
from karp.lex.domain import commands
from karp.lex.domain.value_objects import EntrySchema
from karp.lex.application.use_cases import (
    AddingEntries,
    AddingEntry,
    CreatingEntryRepo,
    CreatingResource,
    DeletingEntry,
    PublishingResource,
    UpdatingEntry,
    UpdatingResource,
)
from karp.lex.application.queries import (
    EntryDto,
    GetEntryDiff,
    GetEntryHistory,
    GetHistory,
    GetHistoryDto,
    GetPublishedResources,
    GetResources,
    HistoryDto,
    ResourceDto,
    ReadOnlyResourceRepository,
    EntryRepoDto,
    ReadOnlyEntryRepoRepositry,
    ListEntryRepos,
)


class Lex(injector.Module):
    @injector.provider
    def entry_uow_factory(self, container: injector.Injector) -> EntryRepositoryUnitOfWorkFactory:
        return InjectorEntryUnitOfWorkRepoFactory(container)

    @injector.provider
    def create_entry_repository(
        self,
        uow: EntryUowRepositoryUnitOfWork,
    ) -> CommandHandler[CreateEntryRepository]:
        return CreatingEntryRepo(
            entry_repo_uow=uow,
        )

    @injector.provider
    def create_resource(
        self,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
        resource_uow: ResourceUnitOfWork,
    ) -> CommandHandler[CreateResource]:
        return CreatingResource(
            entry_repo_uow=entry_repo_uow,
            resource_uow=resource_uow,
        )

    @injector.provider
    def update_resource(
        self,
        resource_uow: ResourceUnitOfWork,
    ) -> CommandHandler[commands.UpdateResource]:
        return UpdatingResource(resource_uow)

    @injector.provider
    def publish_resource(
        self,
        resource_uow: ResourceUnitOfWork,
    ) -> CommandHandler[commands.PublishResource]:
        return PublishingResource(resource_uow)

    @injector.provider
    def add_entry(
        self,
        resource_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork
    ) -> CommandHandler[commands.AddEntry]:
        return AddingEntry(
            resource_uow=resource_uow,
            entry_repo_uow=entry_repo_uow
        )

    @injector.provider
    def add_entries(
        self,
        resource_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork
    ) -> CommandHandler[commands.AddEntries]:
        return AddingEntries(
            resource_uow=resource_uow,
            entry_repo_uow=entry_repo_uow
        )

    @injector.provider
    def update_entry(
        self,
        resource_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork
    ) -> CommandHandler[commands.UpdateEntry]:
        return UpdatingEntry(
            resource_uow=resource_uow,
            entry_repo_uow=entry_repo_uow
        )

    @injector.provider
    def delete_entry(
        self,
        resource_uow: ResourceUnitOfWork,
        entry_repo_uow: EntryUowRepositoryUnitOfWork
    ) -> CommandHandler[commands.DeleteEntry]:
        return DeletingEntry(
            resource_uow=resource_uow,
            entry_repo_uow=entry_repo_uow
        )



"""Pytest entry point."""


import pytest  # pyre-ignore
from sqlalchemy import create_engine
from sqlalchemy.orm import session, sessionmaker

from alembic.config import main as alembic_main

# environ["TESTING"] = "True"
# environ["ELASTICSEARCH_HOST"] = "localhost:9202"
# environ["CONSOLE_LOG_LEVEL"] = "DEBUG"

from karp.db_infrastructure.db import metadata  # nopep8


@pytest.fixture(name="in_memory_sqlite_db")
def fixture_in_memory_sqlite_db():
    engine = create_engine("sqlite:///:memory:")
    metadata.create_all(engine)
    yield engine
    session.close_all_sessions()
    metadata.drop_all(bind=engine)


@pytest.fixture
def sqlite_session_factory(in_memory_sqlite_db):
    yield sessionmaker(bind=in_memory_sqlite_db)




from karp.lex.application.repositories import (
    EntryUowRepositoryUnitOfWork,
)

from . import adapters, factories


class TestCreateEntryRepository:
    def test_create_entry_repository(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd)

        entry_uow_repo_uow = lex_ctx.container.get(
            EntryUowRepositoryUnitOfWork)
        assert entry_uow_repo_uow.was_committed
        assert len(entry_uow_repo_uow.repo) == 1



import abc
import uuid
from typing import Generic, TypeVar, Optional, Type, Union

from karp.foundation.value_objects import unique_id
from .errors import NotFoundError

EntityType = TypeVar("EntityType")


class Repository(Generic[EntityType], abc.ABC):
    EntityNotFound: Type[NotFoundError] = NotFoundError

    def __init__(self):
        self.seen = set()

    def save(self, entity: EntityType):
        self._save(entity)
        self.seen.add(entity)

    @abc.abstractmethod
    def _save(self, entity: EntityType):
        raise NotImplementedError()

    def _check_id_has_correct_type(self, id_) -> None:
        if not isinstance(id_, unique_id.UniqueIdType):
            raise ValueError(
                f"expected UniqueId, got id_ = '{id_}' (type: ´{type(id_)}'"
            )

    def by_id(
        self, id_: unique_id.typing_UniqueId, *, version: Optional[int] = None
    ) -> EntityType:
        self._check_id_has_correct_type(id_)
        entity = self._by_id(id_, version=version)
        if entity:
            self.seen.add(entity)
        else:
            raise self.EntityNotFound(f"Entity with id={id_} is not found")
        return entity

    get_by_id = by_id

    def get_by_id_optional(
        self,
        id_: unique_id.typing_UniqueId,
        *,
        version: Optional[int] = None,
    ) -> Optional[EntityType]:
        self._check_id_has_correct_type(id_)
        entity = self._by_id(id_, version=version)
        if entity:
            self.seen.add(entity)
        return entity

    @abc.abstractmethod
    def _by_id(
        self, id_: Union[uuid.UUID, str], *, version: Optional[int] = None
    ) -> Optional[EntityType]:
        raise NotImplementedError()



from typing import Any, Optional, Set

import attr


class Observer:
    def update(self, **kwargs):
        pass


@attr.s(auto_attribs=True)
class Notifier:
    observers: Set[Observer] = attr.Factory(set)

    def register(self, who: Observer):
        self.observers.add(who)

    def unregister(self, who: Observer):
        try:
            self.observers.remove(who)
        except KeyError:
            pass

    def notify(self, **kwargs):
        for observer in self.observers:
            observer.update(**kwargs)




import collections
import json
import logging
# from karp.infrastructure.unit_of_work import unit_of_work
import sys
import typing
from typing import Dict, List, Optional, Tuple, Iterable

from karp.foundation import (
    events as foundation_events,
    commands as foundation_commands,
)
from karp.lex.domain import events as lex_events
from karp.lex.domain.entities import resource
from karp.lex.application.queries import EntryDto
from karp.search.application.queries import ResourceViews
from karp.search.application.repositories import IndexUnitOfWork
from karp.search.application.transformers import EntryTransformer, PreProcessor
# from .search_service import SearchServiceModule
# import karp.resourcemgr as resourcemgr
# import karp.resourcemgr.entryread as entryread
# from karp.resourcemgr.resource import Resource
from karp import errors as karp_errors
from karp.lex.domain import events, entities
# , errors, events, search_service, model
from karp.search.domain import commands
# from karp.search.domain.search_service import SearchService, IndexEntry
from karp.lex.domain.entities.entry import Entry, create_entry
# from karp.domain.models.resource import Resource
# from karp.domain.repository import ResourceRepository
# from karp.services import context, network_handlers

# from karp.domain.services import network


# from karp.resourcemgr.entrymetadata import EntryMetadata

# search_serviceer = SearchServiceModule()

logger = logging.getLogger("karp")


# def pre_process_resource(
#     resource_obj: Resource,
# ) -> List[Tuple[str, EntryMetadata, Dict]]:
#     metadata = resourcemgr.get_all_metadata(resource_obj)
#     fields = resource_obj.config["fields"].items()
#     entries = resource_obj.entities.query.filter_by(deleted=False)
#     return [
#         (
#             entry.entry_id,
#             metadata[entry.id],
#             transform_to_search_service_entry(resource_obj, json.loads(entry.body), fields),
#         )
#         for entry in entries
#     ]

# metadata = resourcemgr.get_all_metadata(resource_obj)
# fields = resource_obj.config["fields"].items()
# entries = resource_obj.entities.query.filter_by(deleted=False)
# return [
#     (
#         entry.entry_id,
#         metadata[entry.id],
#         transform_to_search_service_entry(resource_obj, json.loads(entry.body), fields),
#     )
#     for entry in entries
# ]


# def research_service(
#     resource_id: str,
#     version: Optional[int] = None,
#     search_entries: Optional[List[Tuple[str, EntryMetadata, Dict]]] = None,
# ) -> None:
#     """
#     If `search_entries` is not given, they will be fetched from DB and processed using `transform_to_search_service_entry`
#     If `search_entries` is given, they most have the same format as the output from `pre_process_resource`
#     """
#     resource_obj = resourcemgr.get_resource(resource_id, version=version)
#     try:
#         search_service_name = search_serviceer.impl.create_search_service(resource_id, resource_obj.config)
#     except NotImplementedError:
#         _logger.error("No SearchService module is loaded. Check your configurations...")
#         sys.exit(errors.NoSearchServiceModuleConfigured)
#     if not search_entries:
#         search_entries = pre_process_resource(resource_obj)
#     add_entries(search_service_name, search_entries, update_refs=False)
#     search_serviceer.impl.publish_search_service(resource_id, search_service_name)
# def research_service(
#     search_serviceer: SearchService,
#     resource_repo: ResourceRepository,
#     resource: Resource,
#     search_entries: Optional[List[IndexEntry]] = None,
# ):


class ReindexingResource(
    foundation_commands.CommandHandler[commands.ReindexResource]
):
    def __init__(
        self,
        index_uow: IndexUnitOfWork,
        resource_views: ResourceViews,
        pre_processor: PreProcessor,
    ) -> None:
        super().__init__()
        self.index_uow = index_uow
        self.resource_views = resource_views
        self.pre_processor = pre_processor

    def execute(
        self,
        cmd: commands.ReindexResource
    ) -> None:
        logger.debug("Reindexing resource '%s'", cmd.resource_id)
        with self.index_uow as uw:
            uw.repo.create_index(
                cmd.resource_id,
                self.resource_views.get_resource_config(cmd.resource_id),
            )
            uw.repo.add_entries(
                cmd.resource_id,
                self.pre_processor.process(cmd.resource_id)
            )
            uw.commit()


def research_service(
    evt: events.ResourcePublished,
):
    print("creating search_service ...")
    search_service_name = search_serviceer.create_search_service(
        resource.resource_id, resource.config)

    if not search_entries:
        print("preprocessing entries ...")
        search_entries = pre_process_resource(
            resource, resource_repo, search_serviceer)
    print(f"adding entries to '{search_service_name}' ...")
    # add_entries(
    #     resource_repo,
    #     search_serviceer,
    #     resource,
    #     search_entries,
    #     search_service_name=search_service_name,
    #     update_refs=False,
    # )
    search_serviceer.add_entries(search_service_name, search_entries)
    print("publishing ...")
    search_serviceer.publish_search_service(
        resource.resource_id, search_service_name)


# def publish_search_service(resource_id: str, version: Optional[int] = None) -> None:
class ResourcePublishedHandler(foundation_events.EventHandler[lex_events.ResourcePublished]):
    def __init__(
        self,
        index_uow: IndexUnitOfWork
    ):
        self.index_uow = index_uow

    def __call__(
        self,
        evt: events.ResourcePublished,
    ) -> None:
        # research_service(evt, ctx)
        with self.index_uow as uw:
            uw.repo.publish_index(evt.resource_id)
            uw.commit()
        # if version:
        #     resourcemgr.publish_resource(resource_id, version)


class CreateSearchServiceHandler(foundation_events.EventHandler[lex_events.ResourceCreated]):
    def __init__(
        self,
        index_uow: IndexUnitOfWork
    ):
        self.index_uow = index_uow

    def collect_new_events(self) -> Iterable[foundation_events.Event]:
        yield from self.index_uow.collect_new_events()

    def __call__(self, evt: events.ResourceCreated):
        with self.index_uow as uw:
            uw.repo.create_index(evt.resource_id, evt.config)
            uw.commit()


# def add_entries(
#     resource_id: str,
#     entries: List[Tuple[str, EntryMetadata, Dict]],
#     update_refs: bool = True,
# ) -> None:
#     search_serviceer.impl.add_entries(resource_id, entries)
#     if update_refs:
#         _update_references(resource_id, [entry_id for (entry_id, _, _) in entries])


class EntryAddedHandler(foundation_events.EventHandler[lex_events.EntryAdded]):
    def __init__(
        self,
        index_uow: IndexUnitOfWork,
        entry_transformer: EntryTransformer,
        resource_views: ResourceViews,
    ):
        self.index_uow = index_uow
        self.entry_transformer = entry_transformer
        self.resource_views = resource_views

    def __call__(
        self,
        evt: events.EntryAdded,
    ):
        with self.index_uow as uw:

            for resource_id in self.resource_views.get_resource_ids(evt.repo_id):
                entry = EntryDto(
                    entity_id=evt.entity_id,
                    entry_id=evt.entry_id,
                    repository_id=evt.repo_id,
                    resource=resource_id,
                    entry=evt.body,
                    message=evt.message,
                    last_modified=evt.timestamp,
                    last_modified_by=evt.user,
                    version=1,
                )
                uw.repo.add_entries(
                    resource_id,
                    [self.entry_transformer.transform(resource_id, entry)]
                )
                self.entry_transformer.update_references(
                    resource_id, [evt.entry_id])
            uw.commit()


class EntryUpdatedHandler(
    foundation_events.EventHandler[lex_events.EntryUpdated]
):
    def __init__(
        self,
        index_uow: IndexUnitOfWork,
        entry_transformer: EntryTransformer,
        resource_views: ResourceViews,
    ):
        self.index_uow = index_uow
        self.entry_transformer = entry_transformer
        self.resource_views = resource_views

    def __call__(
        self,
        evt: events.EntryUpdated,
    ):
        with self.index_uow as uw:

            for resource_id in self.resource_views.get_resource_ids(evt.repo_id):
                entry = EntryDto(
                    entity_id=evt.entity_id,
                    entry_id=evt.entry_id,
                    repository_id=evt.repo_id,
                    resource=resource_id,
                    entry=evt.body,
                    message=evt.message,
                    last_modified=evt.timestamp,
                    last_modified_by=evt.user,
                    version=evt.version,
                )
                uw.repo.add_entries(
                    resource_id,
                    [self.entry_transformer.transform(resource_id, entry)]
                )
                self.entry_transformer.update_references(
                    resource_id, [evt.entry_id])
            uw.commit()
            # add_entries(evt.resource_id, [entry], ctx)
            # ctx.index_uow.commit()

    # def add_entries(
    #     resource_repo: ResourceRepository,
    #     search_serviceer: SearchService,
    #     resource: Resource,
    #     entries: List[Entry],
    #     *,
    #     update_refs: bool = True,
    #     search_service_name: Optional[str] = None,
    # ) -> None:


def add_entries(
    resource_id: str,
    entries: typing.List[entities.Entry],
    *,
    update_refs: bool = True,
    search_service_name: typing.Optional[str] = None,
):
    if not search_service_name:
        search_service_name = resource_id
    with ctx.index_uow:
        with ctx.resource_uow:
            resource = ctx.resource_uow.repo.by_resource_id(resource_id)
            if not resource:
                raise errors.ResourceNotFound(resource_id)
            ctx.index_uow.repo.add_entries(
                search_service_name,
                [transform_to_search_service_entry(resource, entry, ctx)
                 for entry in entries],
            )
            if update_refs:
                _update_references(resource, entries, ctx)
            ctx.resource_uow.commit()
        ctx.index_uow.commit()


class EntryDeletedHandler(
    foundation_events.EventHandler[lex_events.EntryDeleted]
):
    def __init__(
        self,
        index_uow: IndexUnitOfWork,
        entry_transformer: EntryTransformer,
        resource_views: ResourceViews,
    ):
        self.index_uow = index_uow
        self.entry_transformer = entry_transformer
        self.resource_views = resource_views

    def __call__(
        self,
        evt: events.EntryDeleted
    ):
        with self.index_uow as uw:
            for resource_id in self.resource_views.get_resource_ids(evt.repo_id):
                uw.repo.delete_entry(
                    resource_id,
                    entry_id=evt.entry_id
                )
                self.entry_transformer.update_references(
                    resource_id,
                    [evt.entry_id]
                )
            uw.commit()



import collections
import logging
import typing
# from karp.infrastructure.unit_of_work import unit_of_work
from typing import Any, Dict, Iterator, List, Optional, Tuple
from karp.lex.application.queries.entries import EntryDto

from karp.lex.domain import entities
from karp.lex.domain.entities.entry import Entry
from karp.lex.application.queries import GetReferencedEntries, ReferenceDto, ReadOnlyResourceRepository
from karp.lex.application.repositories import EntryUowRepositoryUnitOfWork
# from karp.resourcemgr import get_resource
# import karp.resourcemgr.entryread as entryread
from karp.lex.domain.errors import EntryNotFound
# from karp.services import context

logger = logging.getLogger(__name__)


class GenericGetReferencedEntries(GetReferencedEntries):
    def __init__(
        self,
        resource_repo: ReadOnlyResourceRepository,
        entry_repo_uow: EntryUowRepositoryUnitOfWork,
    ) -> None:
        super().__init__()
        self.resource_repo = resource_repo
        self.entry_repo_uow = entry_repo_uow

    def query(
        self,
        resource_id: str,
        # version: typing.Optional[int],
        entry_id: str,
    ) -> typing.Iterable[ReferenceDto]:
        logger.debug(
            f"network.get_referenced_entries: resource_id={resource_id}")
        logger.debug(f"network.get_referenced_entries: entry_id={entry_id}")
        resource_refs, resource_backrefs = self.get_refs(
            resource_id,  # version=version
        )

        resource = self.resource_repo.get_by_resource_id(resource_id)
        if not resource:
            raise RuntimeError(f"resource '{resource_id}' not found")
        with self.entry_repo_uow:
            entry_uow = self.entry_repo_uow.repo.get_by_id(
                resource.entry_repository_id)
        with entry_uow as uw:
            src_entry = uw.repo.by_entry_id(entry_id,  # version=version
                                            )
            # if not src_entry:
            #     raise EntryNotFoundError(
            #         resource.resource_id, entry_id, resource_version=resource.version
            #     )

        for (
            ref_resource_id,
            ref_resource_version,
            field_name,
            field,
        ) in resource_backrefs:
            other_resource = self.resource_repo.get_by_resource_id(
                ref_resource_id  # , version=version
            )
            if not other_resource:
                logger.info('resource %s not found, skipping ...',
                            ref_resource_id)
                continue
            with self.entry_repo_uow:
                other_entry_uow = self.entry_repo_uow.repo.get_by_id(
                    other_resource.entry_repository_id)
            with other_entry_uow as entries_uw:
                for entry in entries_uw.repo.by_referenceable({field_name: entry_id}):
                    yield _create_ref(ref_resource_id, ref_resource_version, entry)

        # src_body = json.loads(src_entry.body)
        for (ref_resource_id, ref_resource_version, field_name, field) in resource_refs:
            ids = src_entry.body.get(field_name)
            if not field.get("collection", False):
                ids = [ids]
            ref_resource = self.resource_repo.get_by_resource_id(
                ref_resource_id)
            if ref_resource:
                with self.entry_repo_uow.repo.get_by_id(ref_resource.entry_repository_id) as entries_uw:
                    for ref_entry_id in ids:
                        entry = entries_uw.repo.get_by_entry_id_optional(
                            ref_entry_id, version=ref_resource_version
                        )
                        if entry:
                            yield _create_ref(
                                ref_resource_id, ref_resource_version, entry
                            )

    def get_refs(
        self,
        resource_id, version=None
    ) -> Tuple[List[Tuple[str, int, str, Dict]], List[Tuple[str, int, str, Dict]]]:
        """
        Goes through all other resource configs finding resources and fields that refer to this resource
        """
        resource_backrefs = collections.defaultdict(
            lambda: collections.defaultdict(dict))
        resource_refs = collections.defaultdict(
            lambda: collections.defaultdict(dict))

        src_resource = self.resource_repo.get_by_resource_id(
            resource_id, version=version)

        if not src_resource:
            raise RuntimeError(f"Can't find resource '{resource_id}'")
        all_other_resources = [
            resource
            for resource in self.resource_repo.get_published_resources()
            if resource.resource_id != resource_id
        ]

        for field_name, field in src_resource.config["fields"].items():
            if "ref" in field:
                if "resource_id" not in field["ref"]:
                    resource_backrefs[resource_id][version][field_name] = field
                    resource_refs[resource_id][version][field_name] = field
                else:
                    resource_refs[field["ref"]["resource_id"]][
                        field["ref"]["resource_version"]
                    ][field_name] = field
            elif "function" in field and "multi_ref" in field["function"]:
                virtual_field = field["function"]["multi_ref"]
                ref_field = virtual_field["field"]
                if "resource_id" in virtual_field:
                    resource_backrefs[virtual_field["resource_id"]][
                        virtual_field["resource_version"]
                    ][ref_field] = None
                else:
                    resource_backrefs[resource_id][version][ref_field] = None

        for other_resource in all_other_resources:
            # other_resource = get_resource(
            #     resource_def.resource_id, version=resource_def.version
            # )
            for field_name, field in other_resource.config["fields"].items():
                ref = field.get("ref")
                if (
                    ref
                    and ref.get("resource_id") == resource_id
                    and ref.get("resource_version") == version
                ):
                    resource_backrefs[other_resource.resource_id][other_resource.version][
                        field_name
                    ] = field

        def flatten_dict(ref_dict):
            ref_list = []
            for ref_resource_id, versions in ref_dict.items():
                for ref_version, field_names in versions.items():
                    for field_name, field in field_names.items():
                        ref_list.append(
                            (ref_resource_id, ref_version, field_name, field))
            return ref_list

        return flatten_dict(resource_refs), flatten_dict(resource_backrefs)


def _create_ref(
    resource_id: str, resource_version: int, entry: Entry
) -> ReferenceDto:
    return ReferenceDto(
        resource_id=resource_id,
        resource_version=resource_version,
        entry=EntryDto(**entry.dict()),
    )



"""
Query API.
"""
import logging
from typing import List, Optional

from fastapi import (APIRouter, Depends, HTTPException, Path, Query, Security,
                     status)

from karp import errors as karp_errors, auth, search
from karp.search.application.queries import SearchService, QueryRequest
from karp.webapp import schemas

from karp.webapp import dependencies as deps
from karp.webapp.dependencies.fastapi_injector import inject_from_req


logger = logging.getLogger(__name__)


router = APIRouter()


@router.get(
    "/entries/{resource_id}/{entry_ids}",
    description="Returns a list of entries matching the given ids",
    name="Get lexical entries by id",
)
def get_entries_by_id(
    resource_id: str = Path(...,
                            description="The resource to perform operation on"),
    entry_ids: str = Path(
        ...,
        description="Comma-separated. The ids to perform operation on.",
        regex=r"^\w(,\w)*",
    ),
    user: auth.User = Security(deps.get_user_optional, scopes=["read"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    search_service: SearchService = Depends(inject_from_req(SearchService)),
):
    logger.debug("webapp.views.get_entries_by_id")
    if not auth_service.authorize(
        auth.PermissionLevel.read, user, [resource_id]
    ):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="read"'},
        )
    return search_service.search_ids(resource_id, entry_ids)


@router.get("/split/{resources}", name="Query per resource")
def query_split(
    resources: str = Path(
        ...,
        regex=r"^\w+(,\w+)*$",
        description="A comma-separated list of resource identifiers",
    ),
    q: Optional[str] = Query(
        None,
        title="query",
        description="The query. If missing, all entries in chosen resource(s) will be returned.",
    ),
    from_: int = Query(
        0, alias="from", description="Specify which entry should be the first returned."
    ),
    size: int = Query(25, description="Number of entries in page."),
    sort: str = Query(
        None,
        description="The `field` to sort by. If missing, default order for each resource will be used.",
        regex=r"^\w+\|(asc|desc)",
    ),
    lexicon_stats: bool = Query(
        True, description="Show the hit count per lexicon"),
    include_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to return"
    ),
    exclude_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to remove from result"
    ),
    format: schemas.EntryFormat = Query(
        schemas.EntryFormat.json,
        description="Will return the result in the specified format.",
    ),
    user: auth.User = Security(deps.get_user_optional, scopes=["read"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    search_service: SearchService = Depends(inject_from_req(SearchService)),
):
    logger.debug('/query/split called', extra={'resources': resources})
    resource_list = resources.split(",")
    if not auth_service.authorize(
        auth.PermissionLevel.read, user, resource_list
    ):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="read"'},
        )
    query_request = QueryRequest(
        resource_ids=resource_list,
        q=q,
        from_=from_,
        size=size,
        lexicon_stats=lexicon_stats,
    )
    try:
        response = search_service.query_split(query_request)
    except karp_errors.KarpError as err:
        logger.exception(
            "Error occured when calling '/query/split'",
            extra={'resources': resources,
                   'q': q,
                   'error_message': err.message},
        )
        raise
    except search.IncompleteQuery as err:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail={
                'errorCode': karp_errors.ClientErrorCodes.SEARCH_INCOMPLETE_QUERY,
                'message': 'Error in query',
                'failing_query': err.failing_query,
                'error_description': err.error_description,
            }
        )
    return response


@router.get(
    "/{resources}",
    description="Returns a list of entries matching the given query in the given resources. The results are mixed from the given resources.",
    name="Query",
    responses={200: {"content": {"application/json": {}}}},
)
def query(
    resources: str = Path(
        ...,
        regex=r"^\w+(,\w+)*$",
        description="A comma-separated list of resource identifiers",
    ),
    q: Optional[str] = Query(
        None,
        title="query",
        description="The query. If missing, all entries in chosen resource(s) will be returned.",
    ),
    from_: int = Query(
        0, alias="from", description="Specify which entry should be the first returned."
    ),
    size: int = Query(25, description="Number of entries in page."),
    sort: List[str] = Query(
        [],
        description="The `field` to sort by. If missing, default order for each resource will be used.",
        regex=r"^\w+\|(asc|desc)",
    ),
    lexicon_stats: bool = Query(
        True, description="Show the hit count per lexicon"),
    include_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to return"
    ),
    exclude_fields: Optional[List[str]] = Query(
        None, description="Comma-separated list of which fields to remove from result"
    ),
    format_: schemas.EntryFormat = Query(
        schemas.EntryFormat.json,
        alias="format",
        description="Will return the result in the specified format.",
    ),
    user: auth.User = Security(deps.get_user_optional, scopes=["read"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    search_service: SearchService = Depends(inject_from_req(SearchService)),
):
    """
    # Query DSL
    ## Query operators
    - `contains|<field>|<string>` Find all entries where the field <field> contains <string>. More premissive than equals.

    endswith|<field>|<string> Find all entries where the field <field> ends with <string>

    equals|<field>|<string> Find all entries where <field> equals <string>. Stricter than contains

    exists|<field> Find all entries that has the field <field>.

    freetext|<string> Search in all fields for <string> and similar values.

    freergxp|<regex.*> Search in all fields for the regex <regex.*>.

    gt|<field>|<value> Find all entries where <field> is greater than <value>.

    gte|<field>|<value> Find all entries where <field> is greater than or equals <value>.

    lt|<field>|<value> Find all entries where <field> is less than <value>.

    lte|<field>|<value> Find all entries where <field> is less than or equals <value>.

    missing|<field> Search for all entries that doesn't have the field <field>.

    regexp|<field>|<regex.*> Find all entries where the field <field> matches the regex <regex.*>.

    startswith|<field>|<string> Find all entries where <field>starts with <string>.

    Logical Operators
    The logical operators can be used both at top-level and lower-levels.

    not(<expression>) Find all entries that doesn't match the expression <expression>.

    and(<expression1>||<expression2>) Find all entries that matches <expression1> AND <expression2>.

    or(<expression1>||<expression2>) Find all entries that matches <expression1> OR <expression2>.

    Regular expressions
    Always matches complete tokens.
    """
    # Examples
    # not||missing|pos
    # and||freergxp|str.*ng||regexp|pos|str.*ng
    # and||missing|pos||equals|wf||or|blomma|äpple
    # and||equals|wf|sitta||not||equals|wf|satt
    # """
    logger.debug(
        "Called 'query' called with", extra={'resources': resources, 'from': from_, 'size': size}
    )
    resource_list = resources.split(",")
    if not auth_service.authorize(
        auth.PermissionLevel.read, user, resource_list
    ):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="read"'},
        )
    query_request = QueryRequest(
        resource_ids=resource_list,
        q=q,
        from_=from_,
        size=size,
        sort=sort,
        include_fields=include_fields,
        exclude_fields=exclude_fields,
        format_=format_,
        lexicon_stats=lexicon_stats,
    )
    try:
        print(f'{search_service=}')
        response = search_service.query(query_request)

    except karp_errors.KarpError as err:
        logger.exception(
            "Error occured when calling 'query' with",
            extra={'resources': resources,
                   'q': q,
                   'error_message': err.message},
        )
        raise
    return response


def init_app(app):
    app.include_router(router)



"""Model for a lexical entry."""
import enum
import logging
import typing
from typing import Dict, List, Optional, Any

from karp.foundation import constraints
from karp.lex.domain import errors, events
from karp.foundation.entity import TimestampedVersionedEntity
from karp.foundation.value_objects import unique_id

logger = logging.getLogger("karp")


class EntryOp(enum.Enum):
    ADDED = "ADDED"
    DELETED = "DELETED"
    UPDATED = "UPDATED"


class EntryStatus(enum.Enum):
    IN_PROGRESS = "IN-PROGRESS"
    IN_REVIEW = "IN_REVIEW"
    OK = "OK"


class Entry(TimestampedVersionedEntity):
    DiscardedEntityError = errors.DiscardedEntityError

    def __init__(
        self,
        *,
        entry_id: str,
        body: Dict,
        message: str,
        # resource_id: str,
        repository_id: unique_id.UniqueId,
        # IN-PROGRESS, IN-REVIEW, OK, PUBLISHED
        status: EntryStatus = EntryStatus.IN_PROGRESS,
        op: EntryOp = EntryOp.ADDED,
        version: int = 1,
        **kwargs,
        # version: int = 0
    ):
        super().__init__(version=version, **kwargs)
        self._entry_id = entry_id
        self._body = body
        self._op = op
        self._message = "Entry added." if message is None else message
        # self.resource_id = resource_id
        self._status = status
        self._repo_id = repository_id

    @property
    def repo_id(self) -> unique_id.UniqueId:
        return self._repo_id

    @property
    def entry_id(self):
        """The entry_id of this entry."""
        return self._entry_id

    @entry_id.setter
    def entry_id(self, entry_id: str):
        self._check_not_discarded()
        self._entry_id = constraints.length_gt_zero("entry_id", entry_id)

    @property
    def body(self):
        """The body of the entry."""
        return self._body

    @body.setter
    def body(self, body: Dict):
        self._check_not_discarded()
        self._body = body

    @property
    def op(self):
        """The latest operation of this entry."""
        return self._op

    @property
    def status(self):
        """The workflow status of this entry."""
        return self._status

    @status.setter
    def status(self, status: EntryStatus):
        """The workflow status of this entry."""
        self._check_not_discarded()
        self._status = status

    @property
    def message(self):
        """The message for the latest operation of this entry."""
        return self._message

    def dict(self) -> Dict[str, Any]:
        return {
            "entry_id": self._entry_id,
            "entity_id": self.entity_id,
            "resource": "",
            "version": self._version,
            "entry": self._body,
            "last_modified": self._last_modified,
            "last_modified_by": self._last_modified_by,
        }

    def discard(
        self,
        *,
        user: str,
        timestamp: float,
        message: str = None,
    ):
        self._check_not_discarded()
        self._op = EntryOp.DELETED
        self._message = message or "Entry deleted."
        self._discarded = True
        self._last_modified_by = user
        self._last_modified = timestamp
        self._version += 1
        self._record_event(
            events.EntryDeleted(
                entity_id=self.id,
                entry_id=self.entry_id,
                timestamp=self.last_modified,
                user=user,
                message=self._message,
                version=self.version,
                repo_id=self.repo_id,
            )
        )
        # event.mutate(self)
        # event_handler.publish(event)

    def stamp(
        self,
        user: str,
        *,
        message: str = None,
        timestamp: float = None,
        increment_version: bool = True,
    ):
        super().stamp(user, timestamp=timestamp, increment_version=increment_version)
        self._message = message
        self._op = EntryOp.UPDATED
        self._record_event(
            events.EntryUpdated(
                timestamp=self.last_modified,
                entity_id=self.id,
                repo_id=self.repo_id,
                entry_id=self.entry_id,
                body=self.body,
                message=self.message or "",
                user=self.last_modified_by,
                version=self.version,
            )
        )

    def __repr__(self) -> str:
        return f"Entry(id={self._id}, entry_id={self._entry_id}, version={self.version}, last_modified={self._last_modified}, body={self.body})"


# === Factories ===
def create_entry(
    entry_id: str,
    body: Dict,
    *,
    entity_id: unique_id.UniqueId,
    repo_id: unique_id.UniqueId,
    last_modified_by: str = None,
    message: Optional[str] = None,
    last_modified: typing.Optional[float] = None,
) -> Entry:
    if not isinstance(entry_id, str):
        entry_id = str(entry_id)
    entry = Entry(
        entry_id=entry_id,
        body=body,
        message="Entry added." if not message else message,
        status=EntryStatus.IN_PROGRESS,
        op=EntryOp.ADDED,
        # entity_id=unique_id.make_unique_id(),
        version=1,
        last_modified_by="Unknown user" if not last_modified_by else last_modified_by,
        repository_id=repo_id,
        entity_id=entity_id,
        last_modified=last_modified,
    )
    entry.queue_event(
        events.EntryAdded(
            repo_id=repo_id,
            entity_id=entry.id,
            entry_id=entry.entry_id,
            body=entry.body,
            message=entry.message or "",
            user=entry.last_modified_by,
            timestamp=entry.last_modified,
        )
    )
    return entry


# === Repository ===


# class EntryRepositorySettings:
#     """Settings for an EntryRepository."""
#
#     pass
#
#
# @singledispatch
# def create_entry_repository(settings: EntryRepositorySettings) -> EntryRepository:
#     raise RuntimeError(f"Don't know how to handle {settings!r}")



from typing import Callable

import pytest

from karp.lex.domain.events import ResourceCreated

from karp.search.application.repositories import IndexUnitOfWork
from karp.tests.unit.lex import factories as lex_factories
from . import adapters, factories


@pytest.fixture(name="resource_created")
def fixture_resource_created() -> ResourceCreated:
    return lex_factories.ResourceCreatedFactory()


@pytest.mark.parametrize('event_factory,predicate', [
    (None, lambda x: x.created),
    (lex_factories.ResourcePublishedFactory, lambda x: x.published),

])
def test_index_reacts_on_lex_events(
    event_factory,
    predicate: Callable,
    search_unit_ctx: adapters.SearchUnitTestContext,
):
    resource_created = lex_factories.ResourceCreatedFactory()
    search_unit_ctx.event_bus.post(resource_created)
    if event_factory:
        event = event_factory(resource_id=resource_created.resource_id)
        search_unit_ctx.event_bus.post(event)

    index_uow = search_unit_ctx.container.get(
        IndexUnitOfWork)
    assert index_uow.was_committed

    assert predicate(
        index_uow.repo.indicies[resource_created.resource_id])


def test_index_reacts_on_entry_added_event(
    search_unit_ctx: adapters.SearchUnitTestContext,
):
    create_entry_repo = lex_factories.CreateEntryRepositoryFactory()
    search_unit_ctx.command_bus.dispatch(create_entry_repo)
    create_resource = lex_factories.CreateResourceFactory(
        entry_repo_id=create_entry_repo.entity_id)
    search_unit_ctx.command_bus.dispatch(create_resource)
    create_entry = lex_factories.AddEntryFactory(
        resource_id=create_resource.resource_id,
        entry={'baseform': 'bra'},
    )
    search_unit_ctx.command_bus.dispatch(create_entry)
    # if event_factory:
    #     event = event_factory(resource_id=create_resource.resource_id)
    #     search_unit_ctx.event_bus.post(event)

    index_uow = search_unit_ctx.container.get(
        IndexUnitOfWork)
    assert index_uow.was_committed

    assert index_uow.repo.indicies[create_resource.resource_id].created
    assert len(
        index_uow.repo.indicies[create_resource.resource_id].entries) == 1
    assert 'bra' in index_uow.repo.indicies[
        create_resource.resource_id
    ].entries


def test_index_reacts_on_entry_updated_event(
    search_unit_ctx: adapters.SearchUnitTestContext,
):
    create_entry_repo = lex_factories.CreateEntryRepositoryFactory()
    search_unit_ctx.command_bus.dispatch(create_entry_repo)
    create_resource = lex_factories.CreateResourceFactory(
        entry_repo_id=create_entry_repo.entity_id)
    search_unit_ctx.command_bus.dispatch(create_resource)
    create_entry = lex_factories.AddEntryFactory(
        resource_id=create_resource.resource_id,
        entry={'baseform': 'bra'},
    )
    search_unit_ctx.command_bus.dispatch(create_entry)

    update_entry = lex_factories.UpdateEntryFactory(
        resource_id=create_resource.resource_id,
        entry_id='bra',
        entry={'baseform': 'bra', 'wordclass': 'adjektiv'},
        version=1,
    )
    search_unit_ctx.command_bus.dispatch(update_entry)

    index_uow = search_unit_ctx.container.get(
        IndexUnitOfWork)
    assert index_uow.was_committed

    assert index_uow.repo.indicies[create_resource.resource_id].created
    assert len(
        index_uow.repo.indicies[create_resource.resource_id].entries) == 1
    entry = index_uow.repo.indicies[
        create_resource.resource_id
    ].entries['bra']
    assert entry.entry['wordclass'] == 'adjektiv'


def test_index_reacts_on_entry_deleted_event(
    search_unit_ctx: adapters.SearchUnitTestContext,
):
    create_entry_repo = lex_factories.CreateEntryRepositoryFactory()
    search_unit_ctx.command_bus.dispatch(create_entry_repo)
    create_resource = lex_factories.CreateResourceFactory(
        entry_repo_id=create_entry_repo.entity_id)
    search_unit_ctx.command_bus.dispatch(create_resource)
    create_entry = lex_factories.AddEntryFactory(
        resource_id=create_resource.resource_id,
        entry={'baseform': 'bra'},
    )
    search_unit_ctx.command_bus.dispatch(create_entry)

    delete_entry = lex_factories.DeleteEntryFactory(
        resource_id=create_resource.resource_id,
        entry_id='bra',
        version=1,
    )
    search_unit_ctx.command_bus.dispatch(delete_entry)

    index_uow = search_unit_ctx.container.get(
        IndexUnitOfWork)
    assert index_uow.was_committed

    assert index_uow.repo.indicies[create_resource.resource_id].created
    assert len(
        index_uow.repo.indicies[create_resource.resource_id].entries) == 0
    assert 'bra' not in index_uow.repo.indicies[
        create_resource.resource_id
    ].entries


def test_reindex_resource_command(
    search_unit_ctx: adapters.SearchUnitTestContext,
):
    create_entry_repo = lex_factories.CreateEntryRepositoryFactory()
    search_unit_ctx.command_bus.dispatch(create_entry_repo)
    create_resource = lex_factories.CreateResourceFactory(
        entry_repo_id=create_entry_repo.entity_id)
    search_unit_ctx.command_bus.dispatch(create_resource)
    create_entry = lex_factories.AddEntryFactory(
        resource_id=create_resource.resource_id,
        entry={'baseform': 'bra'},
    )
    search_unit_ctx.command_bus.dispatch(create_entry)

    reindex_resource = factories.ReindexResourceFactory(
        resource_id=create_resource.resource_id,
        version=1,
    )
    index_uow = search_unit_ctx.container.get(
        IndexUnitOfWork)
    previous_created_at = index_uow.repo.indicies[create_resource.resource_id].created_at
    search_unit_ctx.command_bus.dispatch(reindex_resource)

    index_uow = search_unit_ctx.container.get(IndexUnitOfWork)
    assert index_uow.was_committed

    assert index_uow.repo.indicies[create_resource.resource_id].created
    assert len(
        index_uow.repo.indicies[create_resource.resource_id].entries) == 1
    assert index_uow.repo.indicies[
        create_resource.resource_id
    ].created_at > previous_created_at


def test_transform_to_index_entry():
    pass



import logging
from os import stat
import traceback
import sys
import time
from typing import Any

try:
    from importlib.metadata import entry_points
except ImportError:
    # used if python < 3.8
    from importlib_metadata import entry_points  # type: ignore

from fastapi import FastAPI, Request, Response, status, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from fastapi.exception_handlers import http_exception_handler
import injector
from sqlalchemy.engine import Connection
from sqlalchemy.orm import Session
import logging
from asgi_correlation_id import CorrelationIdMiddleware
from asgi_correlation_id.context import correlation_id

from karp import main
from karp.foundation import errors as foundation_errors
from karp.foundation.value_objects import unique_id
from karp.auth import errors as auth_errors
from karp.lex.domain import errors as lex_errors
from karp.errors import ClientErrorCodes
from karp.main import modules, config
from karp.webapp.routes import router as api_router
from karp.webapp import tasks
from karp.webapp.contrib import MatomoMiddleware


tags_metadata = [
    {"name": "Querying", "description": "Query"},
    {"name": "Editing"},
    {"name": "Statistics"},
    {"name": "History"},
    {"name": "Resources"},
    {"name": "Health"},
]

logger = logging.getLogger(__name__)


def create_app() -> FastAPI:

    app_context = main.bootstrap_app()

    app = FastAPI(
        title=f"{config.PROJECT_NAME} API",
        redoc_url="/",
        version=config.VERSION,
        openapi_tags=tags_metadata
    )

    app.state.app_context = app_context

    app.add_event_handler(
        "startup",
        tasks.create_start_app_handler(app))
    app.add_event_handler(
        "shutdown",
        tasks.create_stop_app_handler(app)
    )

    main.install_auth_service(
        app_context.container, app_context.settings)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=app_context.settings.get('web.cors.origins', ['*']),
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    app.include_router(api_router)

    modules.load_modules('karp.webapp', app=app)

    from karp.errors import KarpError

    @app.exception_handler(KarpError)
    async def _karp_error_handler(request: Request, exc: KarpError):
        logger.exception(exc)
        return JSONResponse(
            status_code=exc.http_return_code,
            content={"error": exc.message, "errorCode": exc.code},
        )

    @app.exception_handler(foundation_errors.NotFoundError)
    async def _entity_not_found(request: Request, exc: foundation_errors.NotFoundError):
        return JSONResponse(
            status_code=404,
            content={
                'detail': str(exc),
                'type': str(type(exc)),
            }
        )

    @app.exception_handler(auth_errors.ResourceNotFound)
    async def _auth_entity_not_found(request: Request, exc: auth_errors.ResourceNotFound):
        return JSONResponse(
            status_code=404,
            content={
                'detail': str(exc),
            }
        )

    @app.exception_handler(lex_errors.LexDomainError)
    def _lex_error_handler(request: Request, exc: lex_errors.LexDomainError):
        logger.exception(exc)
        return lex_exc2response(exc)

    @app.exception_handler(Exception)
    async def unhandled_exception_handler(request: Request, exc: Exception):
        return await http_exception_handler(
            request,
            HTTPException(
                status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail='Internal server error',
                headers={
                    'X-Request-ID': correlation_id.get() or '',
                    'Access-Control-Expose-Headers': 'X-Request-ID'
                }
            )
        )

    @app.middleware('http')
    async def injector_middleware(request: Request, call_next):

        response: Response = JSONResponse(
            status_code=500,
            content={
                'detail': 'Internal server error'
            }
        )
        try:
            request.state.connection = app_context.container.get(Connection)
            request.state.session = Session(bind=request.state.connection)
            request.state.container = app_context.container.create_child_injector(
                modules.request_configuration(
                    conn=request.state.connection,
                    session=request.state.session,
                ))

            response = await call_next(request)
        finally:
            connection = getattr(request.state, 'connection', None)
            if connection:
                request.state.connection.close()

        return response

    @app.middleware("http")
    async def _logging_middleware(request: Request, call_next) -> Response:
        response: Response = JSONResponse(
            status_code=500,
            content={
                'detail': 'Internal server error'
            }
        )
        start_time = time.time()
        try:
            response = await call_next(request)
        finally:
            process_time = time.time() - start_time
            logger.info(
                "processed a request",
                extra={'status_code': response.status_code,
                       'process_time': process_time, }
            )
        return response

    if app_context.settings['tracking.matomo.url']:
        app.add_middleware(
            MatomoMiddleware,
            idsite=app_context.settings['tracking.matomo.idsite'],
            matomo_url=app_context.settings['tracking.matomo.url'],
            access_token=app_context.settings['tracking.matomo.token'],
        )
    else:
        logger.warning(
            'Tracking to Matomo is not enabled, please set TRACKING_MATOMO_URL.')
    app.add_middleware(CorrelationIdMiddleware)

    return app


def lex_exc2response(exc: lex_errors.LexDomainError) -> JSONResponse:
    status_code = 503
    content: dict[str, Any] = {"message": "Internal server error"}
    if isinstance(exc, lex_errors.UpdateConflict):
        status_code = 400
        content = {
            "message": str(exc),
            "errorCode": ClientErrorCodes.VERSION_CONFLICT,
        }
    return JSONResponse(
        status_code=status_code,
        content=content,
    )



import pytest

import elasticsearch_dsl as es_dsl

from karp.search.domain.query_dsl import (
    KarpQueryV6Parser,
    KarpQueryV6ModelBuilderSemantics,
)
from karp.search_infrastructure.queries.es6_search_service import EsQueryBuilder


@pytest.fixture(scope="session")
def parser() -> KarpQueryV6Parser:
    return KarpQueryV6Parser(
        semantics=KarpQueryV6ModelBuilderSemantics()
    )


@pytest.mark.parametrize("q,expected", [
    ("exists|test", es_dsl.Q('exists', field="test")),
    ("freetext|hej", es_dsl.Q("multi_match", query="hej", fuzziness=1)),
    ("freetext|12", es_dsl.Q("multi_match", query=12)),
    (
        "freergxp|1i.*2",
        es_dsl.Q(
            "query_string",
            query="/1i.*2/",
            default_field="*"
        )
    ),
    (
        "missing|test",
        es_dsl.Q(
            "bool",
            must_not=es_dsl.Q(
                "exists",
                field="test"
            )
        )
    ),
    (
        "startswith|pos|nn",
        es_dsl.Q("regexp", pos="nn.*")
    ),
    (
        "endswith|pos|nn",
        es_dsl.Q("regexp", pos=".*nn")
    ),
    (
        "contains|pos|nn",
        es_dsl.Q("regexp", pos=".*nn.*")
    ),
    (
        "gt|val|lok",
        es_dsl.Q("range", val={"gt": "lok"})
    ),
    (
        "gte|val|2",
        es_dsl.Q("range", val={"gte": 2})
    ),
    (
        "lt|val|lok",
        es_dsl.Q("range", val={"lt": "lok"})
    ),
    (
        "lte|val|lok",
        es_dsl.Q("range", val={"lte": "lok"})
    ),
    (
        "equals|pos|vb",
        es_dsl.Q(
            "match",
            pos={
                "query": "vb",
                "operator": "and"
            }
        )
    ),
    (
        "regexp|kjh|lk.*k",
        es_dsl.Q("regexp", kjh="lk.*k")
    ),
    (
        "not(regexp|kjh|lk.*k)",
        ~es_dsl.Q("regexp", kjh="lk.*k")
    ),
    (
        "and(regexp|baseform|g[oe]t||equals|pos|nn)",
        es_dsl.Q("regexp", baseform="g[oe]t") &
        es_dsl.Q(
            "match",
            pos={
                "query": "nn",
                "operator": "and"
            }
        )
    ),
    (
        "and(regexp|baseform|g[oe]t||equals|pos|nn||regexp|pos|n.*)",
        es_dsl.Q("regexp", baseform="g[oe]t") &
        es_dsl.Q(
            "match",
            pos={
                "query": "nn",
                "operator": "and"
            }
        ) &
        es_dsl.Q("regexp", pos="n.*")
    ),
    (
        "or(regexp|baseform|g[oe]t||equals|pos|nn)",
        es_dsl.Q("regexp", baseform="g[oe]t") |
        es_dsl.Q(
            "match",
            pos={
                "query": "nn",
                "operator": "and"
            }
        )
    ),
    (
        "or(regexp|baseform|g[oe]t||equals|pos|nn||regexp|pos|n.*)",
        es_dsl.Q("regexp", baseform="g[oe]t") |
        es_dsl.Q(
            "match",
            pos={
                "query": "nn",
                "operator": "and"
            }
        ) |
        es_dsl.Q("regexp", pos="n.*")
    ),
    # (
    #     "regexp|or(pos||form)|k.+l",
    #     es_dsl.Q("regexp", pos="k.+l") | es_dsl.Q("regexp", form="k.+l")
    # ),
    # (
    #     "count(baseform)|gt|1",
    #     None,
    # )
])
def test_es_query(parser, q, expected):
    model = parser.parse(q)

    query = EsQueryBuilder().walk(model)

    assert query == expected


@pytest.mark.parametrize("q,expected", [
    (
        "and(exists|pos||exists|vb)",
        es_dsl.Q("bool", must=[
            es_dsl.Q("exists", field="pos"),
            es_dsl.Q("exists", field="vb"),
        ])
    ),
    (
        "or(exists|pos||exists|vb)",
        es_dsl.Q("bool", should=[
            es_dsl.Q("exists", field="pos"),
            es_dsl.Q("exists", field="vb"),
        ])
    ),
    (
        "and(not(exists|pos)||not(exists|vb))",
        es_dsl.Q("bool", must_not=[
            es_dsl.Q("exists", field="pos"),
            es_dsl.Q("exists", field="vb"),
        ])
    ),
#     (
#         "exists|or(pos||vb)",
#         es_dsl
#         )
])
def test_combined_es_query(parser, q, expected):
    query = EsQueryBuilder().walk(parser.parse(q))

    assert query == expected



"""SQL repositories for entries."""
import inspect
import logging
import typing
from typing import Dict, List, Optional, Tuple
from uuid import UUID

import injector
import regex
import sqlalchemy as sa
from sqlalchemy import sql
from sqlalchemy.orm import sessionmaker
import logging

from karp.foundation.value_objects import UniqueId
from karp.foundation.events import EventBus
from karp.lex.domain import errors
from karp.lex.application import repositories
# from karp.domain.errors import NonExistingField, RepositoryError
from karp.lex.domain.entities.entry import (  # EntryRepositorySettings,; EntryRepository,; create_entry_repository,
    Entry, EntryOp, EntryStatus)
from karp.db_infrastructure import db
from karp.lex_infrastructure.sql import sql_models
from karp.db_infrastructure.sql_repository import SqlRepository
from karp.db_infrastructure.sql_unit_of_work import SqlUnitOfWork

logger = logging.getLogger(__name__)

DUPLICATE_PATTERN = r"Duplicate entry '(.+)' for key '(\w+)'"
DUPLICATE_PROG = regex.compile(DUPLICATE_PATTERN)
NO_PROPERTY_PATTERN = regex.compile(r"has no property '(\w+)'")


class SqlEntryRepository(
    repositories.EntryRepository, SqlRepository
):
    def __init__(
        self,
        history_model,
        runtime_model,
        resource_config: Dict,
        # resource_id: str,
        # mapped_class: Any
        *,
        session: db.Session,
    ):
        if not session:
            raise TypeError("session can't be None")
        repositories.EntryRepository.__init__(self)
        SqlRepository.__init__(self, session=session)
        self.history_model = history_model
        self.runtime_model = runtime_model
        self.resource_config = resource_config
        # self.mapped_class = mapped_class
        # self.resource_id = resource_id

    @classmethod
    def from_dict(
        cls,
        name: str,
        resource_config: typing.Dict,
        *,
        session: db.Session,
    ):
        if not session:
            raise TypeError("session can't be None")
        try:
            table_name = name
        except KeyError:
            raise ValueError("Missing 'table_name' in settings.")

        # history_model = db.get_table(table_name)
        # if history_model is None:
        #     history_model = create_history_entry_table(table_name)
        # history_model.create(bind=db.engine, checkfirst=True)

        logger.warning({'table_name': table_name})
        history_model = sql_models.get_or_create_entry_history_model(
            table_name)

        # runtime_table = db.get_table(runtime_table_name)
        # if runtime_table is None:
        #     runtime_table = create_entry_runtime_table(
        #         runtime_table_name, history_model, settings["config"]
        #     )

        if session:
            history_model.__table__.create(bind=session.bind, checkfirst=True)
        # runtime_table.create(bind=db.engine, checkfirst=True)
        runtime_model = sql_models.get_or_create_entry_runtime_model(
            table_name, history_model, resource_config
        )
        if session:
            runtime_model.__table__.create(bind=session.bind, checkfirst=True)
            for child_model in runtime_model.child_tables.values():
                child_model.__table__.create(
                    bind=session.bind, checkfirst=True)
        return cls(
            history_model=history_model,
            runtime_model=runtime_model,
            resource_config=resource_config,
            # resource_id=settings["resource_id"],
            session=session,
        )

    @classmethod
    def _create_repository_settings(
        cls, resource_id: str, resource_config: typing.Dict
    ) -> typing.Dict:
        return {
            "table_name": resource_id,
            "resource_id": resource_id,
        }

    def _save(self, entry: Entry):
        self._check_has_session()

        history_id = self._insert_history(entry)

        runtime_entry_raw = self._entry_to_runtime_dict(history_id, entry)
        try:
            current_db_entry = (
                self._session.query(self.runtime_model)
                .filter_by(entity_id=entry.entity_id)
                .first()
            )

            if not current_db_entry:
                return self._session.add(
                    self.runtime_model(
                        **runtime_entry_raw
                    )
                )

            for key, value in runtime_entry_raw.items():
                setattr(current_db_entry, key, value)
#            update_result = self._session.query(
#                self.runtime_model
#            ).filter_by(
#                id=entry.id
#            ).update(
#                runtime_entry_raw,
#            )
#
#            if update_result != 1:
#                self._session.add(
#                    self.runtime_model(
#                        **runtime_entry_raw
#                    )
#                )
            # return self._session.add(runtime_entry)
        except db.exc.DBAPIError as exc:
            logger.exception('db error')
            raise errors.RepositoryError("db failure") from exc

    def _update(self, entry: Entry):
        self._check_has_session()
        history_id = self._insert_history(entry)

        current_db_entry = (
            self._session.query(self.runtime_model)
            .filter_by(entry_id=entry.entry_id)
            .first()
        )

        if not current_db_entry:
            raise errors.RepositoryError(f"Could not find {entry.entry_id}")

        runtime_dict = self._entry_to_runtime_dict(history_id, entry)
        for key, value in runtime_dict.items():
            setattr(current_db_entry, key, value)

    @classmethod
    def primary_key(cls):
        return "entry_id"

    def move(self, entry: Entry, *, old_entry_id: str):
        self._check_has_session()

        db_entry = (
            self._session.query(self.runtime_model)
            .filter_by(entry_id=old_entry_id)
            .first()
        )
        if not db_entry:
            raise errors.RepositoryError(f"Could not find {entry.entry_id}")
        db_entry.discarded = True

        return self.save(entry)

    def delete(self, entry: Entry):
        self._check_has_session()

        self._insert_history(entry)

        db_entry = (
            self._session.query(self.runtime_model)
            .filter_by(entry_id=entry.entry_id)
            .first()
        )
        if not db_entry:
            raise errors.RepositoryError(f"Could not find {entry.entry_id}")
        db_entry.discarded = True

    def _insert_history(self, entry: Entry):
        self._check_has_session()
        try:
            ins_stmt = db.insert(self.history_model)
            history_dict = self._entry_to_history_dict(entry)
            ins_stmt = ins_stmt.values(**history_dict)
            result = self._session.execute(ins_stmt)
            return result.lastrowid or result.returned_defaults["history_id"]
        except db.exc.DBAPIError as exc:
            raise errors.RepositoryError("db failure") from exc

    def entry_ids(self) -> List[str]:
        self._check_has_session()
        query = self._session.query(
            self.runtime_model).filter_by(discarded=False)
        return [row.entry_id for row in query.all()]
        # return [row.entry_id for row in query.filter_by(discarded=False).all()]

    def _by_entry_id(
        self, entry_id: str, *, version: Optional[int] = None
    ) -> Optional[Entry]:
        self._check_has_session()
        query = self._session.query(self.history_model)
        # query = query.join(
        #     self.runtime_table,
        #     self.history_model.c.history_id == self.runtime_table.c.history_id,
        # )
        query = query.filter_by(entry_id=entry_id)
        if version:
            query = query.filter_by(version=version)
        else:
            query = query.order_by(self.history_model.version.desc())
        row = query.first()
        return self._history_row_to_entry(row) if row else None

    def _by_id(
        self,
        id: str,
        *,
        version: Optional[int] = None,
        after_date: Optional[float] = None,
        before_date: Optional[float] = None,
        oldest_first: bool = False,
    ) -> Optional[Entry]:
        self._check_has_session()
        query = self._session.query(self.history_model)
        query = query.filter_by(entity_id=id)
        if version:
            query = query.filter_by(version=version)
        elif after_date is not None:
            query = query.filter(
                self.history_model.last_modified >= after_date
            ).order_by(self.history_model.last_modified)
        elif before_date:
            query = query.filter(
                self.history_model.last_modified <= before_date
            ).order_by(self.history_model.last_modified.desc())
        elif oldest_first:
            query = query.order_by(self.history_model.last_modified)
        else:
            query = query.order_by(self.history_model.last_modified.desc())
        row = query.first()
        return self._history_row_to_entry(row) if row else None

    def history_by_entry_id(self, entry_id: str) -> List[Entry]:
        self._check_has_session()
        query = self._session.query(self.history_model)
        # query = query.join(
        #     self.runtime_table, self.history_model.c.id == self.runtime_table.c.id
        # )
        return query.filter_by(entry_id=entry_id).all()

    def teardown(self):
        """Use for testing purpose."""
        logger.info("starting teardown")
        #         for child_model in self.runtime_model.child_tables.values():
        #             print(f"droping child_model {child_model} ...")
        #             child_model.__table__.drop(bind=db.engine)
        #         print("droping runtime_model ...")
        #         self.runtime_model.__table__.drop(bind=db.engine)
        logger.info("droping history_model ...")
        self.history_model.__table__.drop(bind=self._session.bind)
        logger.info("dropped history_model")

        # db.metadata.drop_all(
        #     bind=db.engine, tables=[self.runtime_model, self.history_model]
        # )

    def all_entries(self) -> typing.Iterable[Entry]:
        self._check_has_session()

        subq = sql.select(
            self.history_model.entity_id,
            sa.func.max(self.history_model.last_modified).label('maxdate')
        ).group_by(self.history_model.entity_id).subquery('t2')

        stmt = sql.select(self.history_model).join(
            subq,
            sa.and_(
                self.history_model.entity_id == subq.c.entity_id,
                self.history_model.last_modified == subq.c.maxdate,
                self.history_model.discarded == False,
            )
        )
        query = self._session.execute(stmt).scalars()
        # query = self._session.query(
        #     self.history_model).filter_by(discarded=False)
        return [self._history_row_to_entry(db_entry) for db_entry in query.all()]

    def get_total_entries(self) -> int:
        self._check_has_session()
        query = self._session.query(
            self.runtime_model.discarded).filter_by(discarded=False)
        return query.count()

    def by_referenceable(self, filters: Optional[Dict] = None, **kwargs) -> List[Entry]:
        self._check_has_session()
        # query = self._session.query(self.runtime_model)
        query = self._session.query(self.runtime_model, self.history_model).filter(
            self.runtime_model.history_id == self.history_model.history_id
        )
        if filters is None:
            if kwargs is None:
                raise RuntimeError("")
            else:
                filters = kwargs

        joined_filters = []
        simple_filters = {}

        for filter_key in filters.keys():
            # tmp = collections.defaultdict(dict)
            if filter_key in self.resource_config[
                "referenceable"
            ] and self.resource_config["fields"][filter_key].get("collection"):
                logger.debug('collection field', extra={
                             'filter_key': filter_key})
                # child_cls = self.runtime_model.child_tables[filter_key]
                # tmp[child_cls.__tablename__][filter_key] = filters[filter_key]
                # print(f"tmp.values() = {tmp.values()}")
                joined_filters.append({filter_key: filters[filter_key]})
                # query = query.filter(
                #     getattr(self.runtime_model, filter_key).any(filters[filter_key])
                # )
                # query = self._session.query(self.runtime_model.child_tables[filter_key])
                # return query.all()
            else:
                simple_filters[filter_key] = filters[filter_key]
            # joined_filters.extend(tmp.values())

        try:
            query = query.filter_by(**simple_filters)
        except db.exc.InvalidRequestError as exc:
            match = NO_PROPERTY_PATTERN.search(str(exc))
            if match:
                raise errors.NonExistingField(match.group(1)) from exc
            else:
                raise errors.RepositoryError(
                    "Unknown invalid request") from exc

        for child_filters in joined_filters:
            child_table_name = list(child_filters.keys())[0]
            logger.debug('child table name', extra={
                         'table_name': child_table_name})
            child_cls = self.runtime_model.child_tables[child_table_name]
            child_query = self._session.query(
                child_cls).filter_by(**child_filters)
            for child_e in child_query:
                logger.debug('child hit',  child_hit=child_e)
            query = query.join(child_cls).filter_by(**child_filters)
        # result = query.filter_by(**kwargs).all()
        # # query = self._session.query(self.history_model)
        # # query = query.join(
        # #     self.runtime_table,
        # #     self.history_model.c.history_id == self.runtime_table.c.history_id,
        # # )
        # # result = query.filter_by(larger_place=7).all()
        # print(f"result = {result}")
        # return result
        # return query.all()
        return [self._history_row_to_entry(db_entry) for _, db_entry in query.all()]

    def get_history(
        self,
        user_id: Optional[str] = None,
        entry_id: Optional[str] = None,
        from_date: Optional[float] = None,
        to_date: Optional[float] = None,
        from_version: Optional[int] = None,
        to_version: Optional[int] = None,
        offset: int = 0,
        limit: int = 100,
    ):
        self._check_has_session()
        query = self._session.query(self.history_model)
        if user_id:
            query = query.filter_by(last_modified_by=user_id)
        if entry_id:
            query = query.filter_by(entry_id=entry_id)
        if entry_id and from_version:
            query = query.filter(self.history_model.version >= from_version)
        elif from_date is not None:
            query = query.filter(self.history_model.last_modified >= from_date)
        if entry_id and to_version:
            query = query.filter(self.history_model.version < to_version)
        elif to_date is not None:
            query = query.filter(self.history_model.last_modified <= to_date)

        paged_query = query.limit(limit).offset(offset)
        total = query.count()
        return [self._history_row_to_entry(row) for row in paged_query.all()], total

    # def _entry_to_history_row(
    #     self, entry: Entry
    # ) -> Tuple[None, UUID, str, int, float, str, Dict, EntryStatus, str, EntryOp, bool]:
    #     return (
    #         None,  # history_id
    #         entry.id,  # id
    #         entry.entry_id,  # entry_id
    #         entry.version,  # version
    #         entry.last_modified,  # last_modified
    #         entry.last_modified_by,  # last_modified_by
    #         entry.body,  # body
    #         entry.status,  # version
    #         entry.message,  # message
    #         entry.op,  # op
    #         entry.discarded,
    #     )

    def _entry_to_history_dict(
        self, entry: Entry, history_id: Optional[int] = None
    ) -> Dict:
        return {
            "history_id": history_id,
            "entity_id": entry.entity_id,
            "entry_id": entry.entry_id,
            "version": entry.version,
            "last_modified": entry.last_modified,
            "last_modified_by": entry.last_modified_by,
            "body": entry.body,
            "status": entry.status,
            "message": entry.message,
            "op": entry.op,
            "discarded": entry.discarded,
            'repo_id': entry.repo_id,
        }

    def _history_row_to_entry(self, row) -> Entry:
        return Entry(
            body=row.body,
            message=row.message,
            status=row.status,
            op=row.op,
            entity_id=row.entity_id,
            last_modified=row.last_modified,
            last_modified_by=row.last_modified_by,
            discarded=row.discarded,
            version=row.version,
            repository_id=row.repo_id,
            entry_id=row.entry_id,
            # resource_id=self.resource_id,
        )

    def _entry_to_runtime_dict(self, history_id: int, entry: Entry) -> Dict:
        _entry = {
            "entry_id": entry.entry_id,
            "history_id": history_id,
            "entity_id": entry.entity_id,
            "discarded": entry.discarded,
        }
        for field_name in self.resource_config.get("referenceable", ()):
            field_val = entry.body.get(field_name)
            if field_val is None:
                continue
            if self.resource_config["fields"][field_name].get("collection"):
                child_table = self.runtime_model.child_tables[field_name]
                for elem in field_val:
                    if field_name not in _entry:
                        _entry[field_name] = []
                    _entry[field_name].append(
                        child_table(**{field_name: elem}))
            else:
                _entry[field_name] = field_val
        return _entry


class SqlEntryUnitOfWork(
    SqlUnitOfWork,
    repositories.EntryUnitOfWork,
):
    repository_type: str = 'sql_entries_v1'

    def __init__(
        self,
        session_factory: sessionmaker,
        event_bus: EventBus,
        **kwargs,

    ):
        SqlUnitOfWork.__init__(self)
        repositories.EntryUnitOfWork.__init__(
            self, event_bus=event_bus, **kwargs)
        self.session_factory = session_factory
        self._entries = None
        self._session = None

    def _begin(self):
        if self._session is None:
            self._session = self.session_factory()
        if self._entries is None:
            self._entries = SqlEntryRepository.from_dict(
                name=self.name,
                resource_config=self.config,
                session=self._session
            )
        return self

    @property
    def repo(self) -> SqlEntryRepository:
        if self._entries is None:
            raise RuntimeError("No entries")
        return self._entries

    @classmethod
    def from_dict(cls, settings: typing.Dict, resource_config, **kwargs):
        return cls(repo_settings=settings, resource_config=resource_config, **kwargs)

    def collect_new_events(self) -> typing.Iterable:
        if self._entries:
            return super().collect_new_events()
        else:
            return []
# ===== Value objects =====
# class SqlEntryRepositorySettings(EntryRepositorySettings):
#     def __init__(self, *, table_name: str, config: Dict):
#         self.table_name = table_name
#         self.config = config


# @create_entry_repository.register(SqlEntryRepositorySettings)
# def _(settings: SqlEntryRepositorySettings) -> SqlEntryRepository:
#     history_model = sql_models.get_or_create_entry_history_model(settings.table_name)

#     runtime_table_name = f"runtime_{settings.table_name}"

#     runtime_model = sql_models.get_or_create_entry_runtime_model(
#         runtime_table_name, history_model, settings.config
#     )
#     return SqlEntryRepository(history_model, runtime_model, settings.config)

class SqlEntryUowCreator:
    repository_type: str = SqlEntryUnitOfWork.repository_type

    @injector.inject
    def __init__(
        self,
        session_factory: sessionmaker,
        event_bus: EventBus,
    ):
        self._session_factory = session_factory
        self.event_bus = event_bus
        self.cache = {}

    def __call__(
        self,
        entity_id: UniqueId,
        name: str,
        config: Dict,
        connection_str: Optional[str],
        user: str,
        message: str,
        timestamp: float,
    ) -> SqlEntryUnitOfWork:
        if entity_id not in self.cache:
            self.cache[entity_id] = SqlEntryUnitOfWork(
                entity_id=entity_id,
                name=name,
                config=config,
                connection_str=connection_str,
                last_modified_by=user,
                message=message,
                last_modified=timestamp,
                session_factory=self._session_factory,
                event_bus=self.event_bus,
            )
        return self.cache[entity_id]



from .entry_handlers import (
    AddingEntry,
    AddingEntries,
    DeletingEntry,
    UpdatingEntry,
)
from .entry_repo_handlers import CreatingEntryRepo
from .resource_handlers import (
    CreatingResource,
    PublishingResource,
    UpdatingResource,
)



from typing import Dict, Optional

import pytest

from karp.lex.domain import errors

from karp.lex.application import repositories
from karp.lex.application.repositories import ResourceUnitOfWork, EntryUowRepositoryUnitOfWork
from karp.lex.domain import commands

from . import adapters, factories

# from karp.domain import errors, events
# from karp.domain.value_objects.unique_id import make_unique_id
# from karp.lex.domain import commands
# from karp.services import unit_of_work

# from karp.tests.unit.adapters import (FakeEntryUowFactory, FakeResourceUnitOfWork,
#                                       bootstrap_test_app)


class TestAddEntry:
    def test_cannot_add_entry_to_nonexistent_resource(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        with pytest.raises(errors.ResourceNotFound):
            lex_ctx.command_bus.dispatch(
                factories.AddEntryFactory(
                    resource_id="non_existent",
                )
            )

    def test_add_entry(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {
                    "baseform": {
                        "type": "string",
                        "required": True
                    }
                },
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)

        entry_id = "beta"
        cmd3 = factories.AddEntryFactory(
            resource_id=cmd2.resource_id,
            entry={"baseform": entry_id},
        )

        lex_ctx.command_bus.dispatch(cmd3)

        entry_uow_repo_uow = lex_ctx.container.get(
            EntryUowRepositoryUnitOfWork
        )
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)
        entry = uow.repo.by_id(cmd3.entity_id)
        assert entry is not None
        assert entry.id == cmd3.entity_id
        assert entry.entry_id == entry_id
        # assert entry.repo_id == cmd3.resource_id

        assert entry.body == {"baseform": entry_id}
        assert entry.last_modified_by == cmd3.user

        assert uow.was_committed

        # assert (
        #     bus.ctx.index_uow.repo.indicies[resource_id].entries[entry_id].id
        #     == entry_id
        # )
        # assert bus.ctx.index_uow.was_committed

    def test_create_entry_with_same_entry_id_raises(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {
                    "baseform": {
                        "type": "string",
                        "required": True
                    }
                },
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)

        entry_id = "beta"
        cmd3 = factories.AddEntryFactory(
            resource_id=cmd2.resource_id,
            entry={"baseform": entry_id},
        )

        lex_ctx.command_bus.dispatch(cmd3)

        with pytest.raises(errors.IntegrityError):
            lex_ctx.command_bus.dispatch(
                factories.AddEntryFactory(
                    resource_id=cmd3.resource_id,
                    entry={"baseform": entry_id},
                ),
            )


class TestUpdateEntry:
    def test_update_entry(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {
                    "baseform": {
                        "type": "string",
                        "required": True
                    },
                    "a": {
                        "type": "string",
                    }
                },
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)

        entry_id = "beta"
        cmd3 = factories.AddEntryFactory(
            resource_id=cmd2.resource_id,
            entry={"baseform": entry_id, 'a': 'orig'},
        )

        lex_ctx.command_bus.dispatch(cmd3)

        lex_ctx.command_bus.dispatch(
            factories.UpdateEntryFactory(
                entry_id=entry_id,
                version=1,
                resource_id=cmd3.resource_id,
                entry={"baseform": entry_id, "a": "changed", "b": "added"},
            ),
        )

        entry_uow_repo_uow = lex_ctx.container.get(
            EntryUowRepositoryUnitOfWork
        )
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)
        assert uow.was_committed

        entry = uow.repo.by_id(cmd3.entity_id)
        assert entry is not None
        assert entry.body["a"] == "changed"
        assert entry.body["b"] == "added"
        assert entry.version == 2
        # assert (
        #     bus.ctx.index_uow.repo.indicies[resource_id]
        #     .entries[entry.entry_id]
        #     .entry["_entry_version"]
        #     == entry.version
        # )
        # assert bus.ctx.index_uow.was_committed

    def test_cannot_update_entry_in_nonexistent_resource(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        with pytest.raises(errors.ResourceNotFound):
            lex_ctx.command_bus.dispatch(
                commands.UpdateEntry(
                    resource_id="non_existent",
                    entry_id="a",
                    version=3,
                    entry={},
                    user="kristoff@example.com",
                    message="update",
                )
            )

    def test_update_of_entry_id_changes_entry_id(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {
                    "baseform": {
                        "type": "string",
                        "required": True
                    },
                },
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)

        entry_id = "beta"
        cmd3 = factories.AddEntryFactory(
            resource_id=cmd2.resource_id,
            entry={"baseform": entry_id},
        )

        lex_ctx.command_bus.dispatch(cmd3)

        new_entry_id = 'gamma'
        lex_ctx.command_bus.dispatch(
            factories.UpdateEntryFactory(
                entry_id=entry_id,
                version=1,
                resource_id=cmd3.resource_id,
                entry={'baseform': new_entry_id},
            ),
        )

        entry_uow_repo_uow = lex_ctx.container.get(
            EntryUowRepositoryUnitOfWork
        )
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)
        assert uow.was_committed

        entry = uow.repo.by_id(cmd3.entity_id)
        assert entry.entry_id == new_entry_id
        assert entry.version == 2

        entry = uow.repo.by_entry_id(new_entry_id)
        assert entry.id == cmd3.entity_id

        with pytest.raises(errors.EntryNotFound):
            uow.repo.by_entry_id(entry_id)


class TestDeleteEntry:
    def test_can_delete_entry(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {
                    "baseform": {
                        "type": "string",
                        "required": True
                    },
                    "a": {
                        "type": "string",
                    }
                },
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)

        entry_id = "beta"
        cmd3 = factories.AddEntryFactory(
            resource_id=cmd2.resource_id,
            entry={"baseform": entry_id, 'a': 'orig'},
        )

        lex_ctx.command_bus.dispatch(cmd3)

        lex_ctx.command_bus.dispatch(
            commands.DeleteEntry(
                entry_id=entry_id,
                version=1,
                resource_id=cmd3.resource_id,
                message="deleted",
                user="bob",
            ),
        )

        entry_uow_repo_uow = lex_ctx.container.get(
            EntryUowRepositoryUnitOfWork
        )
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)
        assert uow.was_committed

        entry = uow.repo.by_id(cmd3.entity_id)
        assert entry.version == 2
        assert entry.discarded

        entry = uow.repo.by_entry_id(entry_id)
        assert entry.version == 2
        assert entry.discarded
        # assert (
        #     entry.entry_id not in bus.ctx.index_uow.repo.indicies[resource_id].entries
        # )
        # assert bus.ctx.index_uow.was_committed



from typing import Iterable, Optional

import sqlalchemy as sa
from sqlalchemy import sql

from karp.lex.application.queries import GetPublishedResources, ResourceDto, GetResources
from karp.lex.application.queries.resources import ReadOnlyResourceRepository
from karp.lex.domain.entities import resource
from karp.lex_infrastructure.sql.sql_models import ResourceModel
from karp.lex_infrastructure.queries.base import SqlQuery


class SqlGetPublishedResources(
    GetPublishedResources,
    SqlQuery
):
    def query(self) -> Iterable[ResourceDto]:
        subq = sql.select(
            ResourceModel.entity_id,
            sa.func.max(ResourceModel.last_modified).label('maxdate')
        ).group_by(ResourceModel.entity_id).subquery('t2')

        stmt = sql.select(ResourceModel).join(
            subq,
            sa.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
                ResourceModel.is_published == True
            )
        )
        return (
            _row_to_dto(row)
            for row in self._conn.execute(stmt)
        )


class SqlGetResources(
    GetResources,
    SqlQuery
):
    def query(self) -> Iterable[ResourceDto]:
        subq = sql.select(
            ResourceModel.entity_id,
            sa.func.max(ResourceModel.last_modified).label('maxdate')
        ).group_by(ResourceModel.entity_id).subquery('t2')

        stmt = sql.select(ResourceModel).join(
            subq,
            sa.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
            )
        )
        return (
            _row_to_dto(row)
            for row in self._conn.execute(stmt)
        )


class SqlReadOnlyResourceRepository(
    ReadOnlyResourceRepository,
    SqlQuery
):
    def get_by_resource_id(
        self,
        resource_id: str,
        version: Optional[int] = None
    ) -> Optional[ResourceDto]:
        filters: dict[str, str | int] = {
            'resource_id': resource_id
        }
        if version:
            filters['version'] = version
        stmt = sql.select(
            ResourceModel
        ).filter_by(**filters).order_by(
            ResourceModel.last_modified.desc()
        )
        row = self._conn.execute(stmt).first()

        return _row_to_dto(row) if row else None

    def get_published_resources(self) -> Iterable[ResourceDto]:
        subq = sql.select(
            ResourceModel.entity_id,
            sa.func.max(ResourceModel.last_modified).label('maxdate')
        ).group_by(ResourceModel.entity_id).subquery('t2')

        stmt = sql.select(ResourceModel).join(
            subq,
            sa.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
                ResourceModel.is_published == True
            )
        )
        return (
            _row_to_dto(row)
            for row in self._conn.execute(stmt)
        )

    def get_all_resources(self) -> Iterable[ResourceDto]:
        subq = sql.select(
            ResourceModel.entity_id,
            sa.func.max(ResourceModel.last_modified).label('maxdate')
        ).group_by(ResourceModel.entity_id).subquery('t2')

        stmt = sql.select(ResourceModel).join(
            subq,
            sa.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
            )
        )
        return (
            _row_to_dto(row)
            for row in self._conn.execute(stmt)
        )


def _row_to_dto(row_proxy) -> ResourceDto:
    return ResourceDto(
        entity_id=row_proxy.entity_id,
        resource_id=row_proxy.resource_id,
        version=row_proxy.version,
        config=row_proxy.config,
        is_published=row_proxy.is_published,
        entry_repository_id=row_proxy.entry_repo_id,
        last_modified=row_proxy.last_modified,
        last_modified_by=row_proxy.last_modified_by,
        message=row_proxy.message,
        name=row_proxy.name,
        discarded=row_proxy.discarded,
    )



import pytest
from starlette import status

from karp import auth
from karp.foundation.value_objects import unique_id
from karp.webapp.schemas import ResourceCreate, ResourcePublic


@pytest.fixture
def new_resource() -> ResourceCreate:
    return ResourceCreate(
        resource_id='test_resource',
        name='Test resource',
        message='test',
        config={
            'fields': {
                'foo': {'type': 'string'}
            },
            'id': 'foo',
        },
    )


class TestResourcesRoutes:
    def test_get_resources_exist(self, fa_data_client):
        response = fa_data_client.get('/resources')
        assert response.status_code != status.HTTP_404_NOT_FOUND

    def test_post_resources_exist(self, fa_data_client):
        response = fa_data_client.post("/resources")
        assert response.status_code != status.HTTP_404_NOT_FOUND

    def test_get_resource_permissionss_exist(self, fa_data_client):
        response = fa_data_client.get('/resources/permissions')
        assert response.status_code != status.HTTP_404_NOT_FOUND

    def test_get_resources(self, fa_data_client):
        response = fa_data_client.get('/resources/permissions')

        assert response.status_code == status.HTTP_200_OK

        response_data = response.json()

        assert len(response_data) == 2
        assert response_data[0] == {
            'resource_id': 'places',
            'protected': None,
        }
        assert response_data[1] == {
            'resource_id': 'municipalities',
            'protected': 'READ',
        }


class TestCreateResource:
    def test_missing_auth_header_returns_403(self, fa_data_client):
        response = fa_data_client.post("/resources/", json={})
        assert response.status_code == status.HTTP_403_FORBIDDEN

    def test_invalid_data_returns_422(
        self,
        fa_data_client,
        admin_token: auth.AccessToken
    ):
        response = fa_data_client.post(
            "/resources/",
            json={},
            headers=admin_token.as_header(),
        )
        print(f'{response.json()=}')
        assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY

    def test_valid_input_creates_resource(
        self,
        fa_data_client,
        new_resource: ResourceCreate,
        admin_token: auth.AccessToken
    ):
        response = fa_data_client.post(
            '/resources/',
            json=new_resource.dict(),
            headers=admin_token.as_header(),
        )
        print(f'{response.json()=}')
        assert response.status_code == status.HTTP_201_CREATED

        created_resource = ResourceCreate(
            **response.json(),
        )
        assert created_resource == new_resource


class TestGetResource:
    def test_get_resource_by_resource_id(self, fa_data_client):
        response = fa_data_client.get('/resources/test_resource')
        print(f'{response.json()=}')
        assert response.status_code == status.HTTP_200_OK

        resource = ResourcePublic(**response.json())
        assert resource.resource_id == 'test_resource'




from datetime import datetime
import logging
import re
from typing import Dict, List, Optional, Any, Tuple, Union

import elasticsearch
import logging

from karp.foundation.events import EventBus

from karp.lex.domain.entities import Entry
from karp.search.application.repositories import (
    Index,
    IndexEntry,
    IndexUnitOfWork,
)


logger = logging.getLogger(__name__)

KARP_CONFIGINDEX = "karp_config"
KARP_CONFIGINDEX_TYPE = "configs"


class Es6Index(Index):
    def __init__(self, es: elasticsearch.Elasticsearch):
        self.es: elasticsearch.Elasticsearch = es
        if not self.es.indices.exists(index=KARP_CONFIGINDEX):
            self.es.indices.create(
                index=KARP_CONFIGINDEX,
                body={
                    "settings": {
                        "number_of_shards": 1,
                        "number_of_replicas": 1,
                        "refresh_interval": -1,
                    },
                    "mappings": {
                        KARP_CONFIGINDEX_TYPE: {
                            "dynamic": False,
                            "properties": {"index_name": {"type": "text"}},
                        }
                    },
                },
            )
        analyzed_fields, sortable_fields = self._init_field_mapping()
        self.analyzed_fields: Dict[str, List[str]] = analyzed_fields
        self.sortable_fields: Dict[str, Dict[str, List[str]]] = sortable_fields

    @property
    def seen(self):
        return []

    def create_index(self, resource_id, config):
        logger.info("creating es mapping")
        mapping = create_es6_mapping(config)

        settings = {
            "number_of_shards": 1,
            "number_of_replicas": 1,
            "refresh_interval": -1,
        }
        if 'settings' in mapping:
            settings.update(mapping['settings'])
            del mapping['settings']
        properties = mapping["properties"]
        properties["freetext"] = {"type": "text"}
        disabled_property = {"enabled": False}
        properties["_entry_version"] = disabled_property
        properties["_last_modified"] = disabled_property
        properties["_last_modified_by"] = disabled_property

        body = {
            "settings": settings,
            "mappings": {"entry": mapping},
        }

        date = datetime.now().strftime("%Y-%m-%d-%H%M%S%f")
        index_name = resource_id + "_" + date
        logger.info('creating index', extra={
                    'index_name': index_name, 'body': body})
        result = self.es.indices.create(index=index_name, body=body)
        if "error" in result:
            logger.error('failed to create index',
                         extra={'index_name': index_name, 'body': body})
            raise RuntimeError("failed to create index")
        logger.info("index created")
        self._set_index_name_for_resource(resource_id, index_name)
        return index_name

    def _set_index_name_for_resource(self, resource_id: str, index_name: str) -> str:
        self.es.index(
            index=KARP_CONFIGINDEX,
            id=resource_id,
            doc_type=KARP_CONFIGINDEX_TYPE,
            body={"index_name": index_name},
        )
        return index_name

    def _get_index_name_for_resource(self, resource_id: str) -> str:
        try:
            res = self.es.get(
                index=KARP_CONFIGINDEX, id=resource_id, doc_type=KARP_CONFIGINDEX_TYPE
            )
        except es_exceptions.NotFoundError as err:
            logger.debug(
                "didn't find index_name for resource '%s' details: %s", resource_id, err)
            return self._set_index_name_for_resource(resource_id, resource_id)
        return res["_source"]["index_name"]

    def publish_index(self, resource_id: str):
        if self.es.indices.exists_alias(name=resource_id):
            self.es.indices.delete_alias(name=resource_id, index="*")

        index_name = self._get_index_name_for_resource(resource_id)
        self.on_publish_resource(resource_id, index_name)
        logger.info('publishing resource',
                    extra={'resource_id': resource_id, 'index_name': index_name})
        self.es.indices.put_alias(name=resource_id, index=index_name)

    def add_entries(self, resource_id: str, entries: List[IndexEntry]):
        index_name = self._get_index_name_for_resource(resource_id)
        index_to_es = []
        for entry in entries:
            assert isinstance(entry, IndexEntry)
            # entry.update(metadata.to_dict())
            index_to_es.append(
                {
                    "_index": index_name,
                    "_id": entry.id,
                    "_type": "entry",
                    "_source": entry.entry,
                }
            )

        elasticsearch.helpers.bulk(self.es, index_to_es, refresh=True)

    def delete_entry(
        self,
        resource_id: str,
        *,
        entry_id: Optional[str] = None,
        entry: Optional[Entry] = None,
    ):
        if not entry and not entry_id:
            raise ValueError("Must give either 'entry' or 'entry_id'.")
        if entry:
            entry_id = entry.entry_id
        logger.info('deleting entry', extra={'entry_id': entry_id,
                    'resource_id': resource_id})
        index_name = self._get_index_name_for_resource(resource_id)
        try:
            self.es.delete(
                index=index_name,
                doc_type="entry",
                id=entry_id,
                refresh=True,
            )
        except elasticsearch.exceptions.ElasticsearchException:
            logger.exception('Error deleting entry',
                             extra={'entry_id': entry_id, 'resource_id': resource_id, 'index_name': index_name, })

    @staticmethod
    def get_analyzed_fields_from_mapping(
        properties: Dict[str, Dict[str, Dict[str, Any]]]
    ) -> List[str]:
        analyzed_fields = []

        for prop_name, prop_values in properties.items():
            if "properties" in prop_values:
                res = Es6Index.get_analyzed_fields_from_mapping(
                    prop_values["properties"]
                )
                analyzed_fields.extend(
                    [prop_name + "." + prop for prop in res])
            else:
                if prop_values["type"] == "text":
                    analyzed_fields.append(prop_name)
        return analyzed_fields

    def _init_field_mapping(
        self,
    ) -> Tuple[Dict[str, List[str]], Dict[str, Dict[str, List[str]]]]:
        """
        Create a field mapping based on the mappings of elasticsearch
        currently the only information we need is if a field is analyzed (i.e. text)
        or not.
        """

        field_mapping: Dict[str, List[str]] = {}
        sortable_fields = {}
        # Doesn't work for tests, can't find resource_definition
        # for resource in resourcemgr.get_available_resources():
        #     mapping = self.es.indices.get_mapping(index=resource.resource_id)
        #     field_mapping[resource.resource_id] = parse_mapping(
        #         next(iter(mapping.values()))['mappings']['entry']['properties']
        #     )
        aliases = self._get_all_aliases()
        mapping: Dict[
            str, Dict[str, Dict[str, Dict[str, Dict]]]
        ] = self.es.indices.get_mapping()
        # print(f"mapping = {mapping}")
        for (alias, index) in aliases:
            if (
                "mappings" in mapping[index]
                and "entry" in mapping[index]["mappings"]
                and "properties" in mapping[index]["mappings"]["entry"]
            ):
                field_mapping[alias] = Es6Index.get_analyzed_fields_from_mapping(
                    mapping[index]["mappings"]["entry"]["properties"]
                )
                sortable_fields[alias] = Es6Index.create_sortable_map_from_mapping(
                    mapping[index]["mappings"]["entry"]["properties"]
                )
        return field_mapping, sortable_fields

    def _get_index_mappings(
        self, index: Optional[str] = None
    ) -> Dict[str, Dict[str, Dict[str, Dict[str, Dict]]]]:
        kwargs = {"index": index} if index is not None else {}
        return self.es.indices.get_mapping(**kwargs)

    def _get_all_aliases(self) -> List[Tuple[str, str]]:
        """
        :return: a list of tuples (alias_name, index_name)
        """
        result = self.es.cat.aliases(h="alias,index")
        print(f"{result}")
        index_names = []
        for index_name in result.split("\n")[:-1]:
            print(f"index_name = {index_name}")
            if index_name[0] != ".":
                match = re.search(r"([^ ]*) +(.*)", index_name)
                if match:
                    groups = match.groups()
                    alias = groups[0]
                    index = groups[1]
                    index_names.append((alias, index))
        return index_names

    def translate_sort_fields(
        self, resources: List[str], sort_values: List[str]
    ) -> List[Union[str, Dict[str, Dict[str, str]]]]:
        """Translate sort field to ES sort fields.

        Arguments:
            sort_values {List[str]} -- values to sort by

        Returns:
            List[str] -- values that ES can sort by.
        """
        translated_sort_fields: List[Union[str,
                                           Dict[str, Dict[str, str]]]] = []
        for sort_value in sort_values:
            sort_order = None
            if "|" in sort_value:
                sort_value, sort_order = sort_value.split("|", 1)
            for resource_id in resources:
                if sort_order:
                    translated_sort_fields.extend(
                        (
                            {field: {"order": sort_order}}
                            for field in self.translate_sort_field(
                                resource_id, sort_value
                            )
                        )
                    )
                translated_sort_fields.extend(
                    self.translate_sort_field(resource_id, sort_value)
                )

        return translated_sort_fields

    def translate_sort_field(self, resource_id: str, sort_value: str) -> List[str]:
        print(
            f"es6_indextranslate_sort_field: sortable_fields[{resource_id}] = {self.sortable_fields[resource_id]}"
        )
        if sort_value in self.sortable_fields[resource_id]:
            return self.sortable_fields[resource_id][sort_value]
        else:
            raise UnsupportedField(
                f"You can't sort by field '{sort_value}' for resource '{resource_id}'"
            )

    def on_publish_resource(self, alias_name: str, index_name: str):
        mapping = self._get_index_mappings(index=index_name)
        if (
            "mappings" in mapping[index_name]
            and "entry" in mapping[index_name]["mappings"]
            and "properties" in mapping[index_name]["mappings"]["entry"]
        ):
            self.analyzed_fields[
                alias_name
            ] = Es6Index.get_analyzed_fields_from_mapping(
                mapping[index_name]["mappings"]["entry"]["properties"]
            )
            self.sortable_fields[
                alias_name
            ] = Es6Index.create_sortable_map_from_mapping(
                mapping[index_name]["mappings"]["entry"]["properties"]
            )

    @staticmethod
    def create_sortable_map_from_mapping(properties: Dict) -> Dict[str, List[str]]:
        sortable_map = {}

        def parse_prop_value(sort_map, base_name, prop_name, prop_value: Dict):
            if "properties" in prop_value:
                for ext_name, ext_value in prop_value["properties"].items():
                    ext_base_name = f"{base_name}.{ext_name}"
                    parse_prop_value(sort_map, ext_base_name,
                                     ext_base_name, ext_value)
                return
            if prop_value["type"] in [
                "boolean",
                "date",
                "double",
                "keyword",
                "long",
                "ip",
            ]:
                sort_map[base_name] = [prop_name]
                sort_map[prop_name] = [prop_name]
                return
            if prop_value["type"] == "text":
                if "fields" in prop_value:
                    for ext_name, ext_value in prop_value["fields"].items():
                        parse_prop_value(
                            sort_map, base_name, f"{base_name}.{ext_name}", ext_value
                        )
                return

        for prop_name, prop_value in properties.items():
            parse_prop_value(sortable_map, prop_name, prop_name, prop_value)
            # if prop_value["type"] in ["boolean", "date", "double", "keyword", "long", "ip"]:
            #     sortable_map[prop_name] = prop_name
            # if prop_value["type"] == "text":
            #     if "fields" in prop_value:

        return sortable_map


# def parse_sortable_fields(properties: Dict[str, Any]) -> Dict[str, List[str]]:
#     for prop_name, prop_value in properties.items():
#         if prop_value["type"] in ["boolean", "date", "double", "keyword", "long", "ip"]:
#             return [prop_name]


def _create_es_mapping(config):
    es_mapping = {"dynamic": False, "properties": {}}

    fields = config["fields"]

    def recursive_field(parent_schema, parent_field_name, parent_field_def):
        if parent_field_def.get("virtual", False):
            fun = parent_field_def["function"]
            if list(fun.keys())[0] == "multi_ref":
                res_object = fun["multi_ref"]["result"]
                recursive_field(parent_schema, "v_" +
                                parent_field_name, res_object)
            if "result" in fun:
                res_object = fun["result"]
                recursive_field(parent_schema, "v_" +
                                parent_field_name, res_object)
            return
        if parent_field_def.get("ref"):
            if "field" in parent_field_def["ref"]:
                res_object = parent_field_def["ref"]["field"]
            else:
                res_object = {}
                res_object.update(parent_field_def)
                del res_object["ref"]
            recursive_field(parent_schema, "v_" +
                            parent_field_name, res_object)
        if parent_field_def["type"] != "object":
            # TODO this will not work when we have user defined types, s.a. saldoid
            # TODO number can be float/non-float, strings can be keyword or text in need of analyzing etc.
            if parent_field_def["type"] == "integer":
                mapped_type = "long"
            elif parent_field_def["type"] == "number":
                mapped_type = "double"
            elif parent_field_def["type"] == "boolean":
                mapped_type = "boolean"
            elif parent_field_def["type"] == "string":
                mapped_type = "text"
            elif parent_field_def['type'] == 'long_string':
                mapped_type = 'text'
            else:
                mapped_type = "keyword"
            result = {"type": mapped_type}
            if parent_field_def['type'] == 'string':
                result["fields"] = {"raw": {"type": "keyword"}}
        else:
            result = {"properties": {}}

            for child_field_name, child_field_def in parent_field_def["fields"].items():
                recursive_field(result, child_field_name, child_field_def)

        parent_schema["properties"][parent_field_name] = result

    for field_name, field_def in fields.items():
        print(f"creating mapping for field '{field_name}'")
        recursive_field(es_mapping, field_name, field_def)

    return es_mapping


def create_es6_mapping(config: Dict) -> Dict:
    mapping = _create_es_mapping(config)
    mapping['settings'] = {
        'analysis': {
            'analyzer': {
                'default': {
                    'char_filter': [
                        'compound',
                        'swedish_aa',
                        'swedish_ae',
                        'swedish_oe'
                    ],
                    'filter': [
                        'swedish_folding',
                        'lowercase'
                    ],
                    'tokenizer': 'standard'
                }
            },
            'char_filter': {
                'compound': {
                    'pattern': '-',
                    'replacement': '',
                    'type': 'pattern_replace'
                },
                'swedish_aa': {
                    'pattern': '[Ǻǻ]',
                    'replacement': 'å',
                    'type': 'pattern_replace'
                },
                'swedish_ae': {
                    'pattern': '[æÆǞǟ]',
                    'replacement': 'ä',
                    'type': 'pattern_replace'
                },
                'swedish_oe': {
                    'pattern': '[ØøŒœØ̈ø̈ȪȫŐőÕõṌṍṎṏȬȭǾǿǬǭŌōṒṓṐṑ]',
                    'replacement': 'ö',
                    'type': 'pattern_replace'
                },
            },
            'filter': {
                'swedish_folding': {
                    'type': 'icu_folding',
                    'unicodeSetFilter': '[^åäöÅÄÖ]'
                },
                'swedish_sort': {
                    'language': 'sv',
                    'type': 'icu_collation'
                }
            }
        }
    }
    return mapping


class Es6IndexUnitOfWork(
    IndexUnitOfWork
):
    def __init__(
        self,
        es: elasticsearch.Elasticsearch,
        event_bus: EventBus,
    ) -> None:
        super().__init__(event_bus=event_bus)
        self._index = Es6Index(es=es)

    # @classmethod
    # def from_dict(cls, **kwargs):
    #     return cls()

    def _commit(self):
        logger.debug('Calling _commit in Es6IndexUnitOfWork')

    def rollback(self):
        return super().rollback()

    @property
    def repo(self) -> Es6Index:
        return self._index

    def _close(self):
        pass



import logging

from slack_log_utils import SlackWebhookFormatter, SlackWebhookHandler


def get_slack_logging_handler(secret):
    slack_formatter = SlackWebhookFormatter()
    slack_handler = SlackWebhookHandler(url=secret, level=logging.ERROR)
    slack_handler.setFormatter(slack_formatter)
    return slack_handler



import dataclasses
from typing import Dict, List, Optional

import injector

from karp.foundation.commands import CommandBus
from karp.foundation.events import EventBus
from karp.foundation.time import utc_now
from karp.search.application.repositories import IndexUnitOfWork, Index, IndexEntry
from karp.tests.foundation.adapters import FakeUnitOfWork


@dataclasses.dataclass
class SearchUnitTestContext:
    container: injector.Injector
    command_bus: CommandBus
    event_bus: EventBus


class FakeIndex(Index):
    @dataclasses.dataclass
    class Index:
        config: Dict
        created_at: float
        entries: Dict[str, IndexEntry] = dataclasses.field(
            default_factory=dict
        )
        created: bool = True
        published: bool = False

    def __init__(self) -> None:
        super().__init__()
        self.indicies = {}
        self.seen = []

    def create_index(self, resource_id: str, config: Dict):
        self.indicies[resource_id] = FakeIndex.Index(
            config=config, created_at=utc_now())

    def publish_index(self, alias_name: str, index_name: str = None):
        self.indicies[alias_name].published = True

    def add_entries(self, resource_id: str, entries: List[IndexEntry]):
        for entry in entries:
            self.indicies[resource_id].entries[entry.id] = entry

    def delete_entry(
        self,
        resource_id: str,
        *,
        entry_id: Optional[str],
        # entry: typing.Optional[model.Entry]
    ):
        del self.indicies[resource_id].entries[entry_id]

#    def search_ids(self, resource_id: str, entry_ids: str):
#        return {}
#
#    def query(self, request: search_service.QueryRequest):
#        return {}
#
#    def query_split(self, request: search_service.QueryRequest):
#        return {}
#
#    def statistics(self, resource_id: str, field: str):
#        return {}


class FakeIndexUnitOfWork(
    FakeUnitOfWork, IndexUnitOfWork
):
    def __init__(self):
        # super().__init__()
        self._index = FakeIndex()

    @property
    def repo(self) -> FakeIndex:
        return self._index


class FakeSearchInfrastructure(injector.Module):
    @injector.provider
    @injector.singleton
    def index_uow(self) -> IndexUnitOfWork:
        return FakeIndexUnitOfWork()



import pytest

from typer import Typer
from typer.testing import CliRunner
from karp import lex

from karp.main import AppContext


class TestCliResourceLifetime:
    def test_help(self, runner: CliRunner, cliapp: Typer):
        result = runner.invoke(cliapp, ['resource', '--help'])
        assert result.exit_code == 0

    def test_create_and_publish_repo(
        self,
        runner: CliRunner,
        cliapp: Typer,
        app_context: AppContext,
    ):
        result = runner.invoke(
            cliapp, ['entry-repo', 'create', 'karp/tests/data/config/lexlex.json'])
        print(f'{result.stdout=}')
        assert result.exit_code == 0

        entry_repo_repo = app_context.container.get(
            lex.ReadOnlyEntryRepoRepositry)

        entry_repo = entry_repo_repo.get_by_name('lexlex')
        assert entry_repo is not None

        result = runner.invoke(
            cliapp, ['resource', 'create',
                     'karp/tests/data/config/lexlex.json', '--entry-repo-id', str(entry_repo.entity_id)]
        )
        print(f'{result.stdout=}')
        assert result.exit_code == 0

        resource_repo = app_context.container.get(
            lex.ReadOnlyResourceRepository)

        assert resource_repo.get_by_resource_id('lexlex') is not None



"""Commands top-level module."""

from .entry_commands import (
    AddEntries,
    AddEntry,
    DeleteEntry,
    UpdateEntry,
)
from .entry_repo_commands import CreateEntryRepository
from .resource_commands import (CreateResource, PublishResource, UpdateResource)
